#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import re


__all__ = ["new"]


SHEBANG = """#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
TOOL_METADATA = """__author__ = "Your name"
__version__ = "1.0"
__copyright__ = "AGPLv3 (http://www.gnu.org/licenses/agpl.html)"
#__reference__ = ""
#__source__ = ""
#__training__ = ""
# TODO: complete the docstring
__doc__ = \"\"\"
This tool ...
\"\"\"
# TODO: complete examples
__examples__ = [""]

"""
TOOL_SECTIONS = """# --------------------- IMPORTS SECTION ---------------------
{imports}# TODO: complete imports
# non-standard imports with warning if dependencies are missing
try:
    # TODO: special imports
except ImportError:
    print("Missing dependencies, please run 'sudo pip install ...'")
    sys.exit(1)


# -------------------- CONSTANTS SECTION --------------------
# TODO: define constants here


# -------------------- FUNCTIONS SECTION --------------------
# TODO: define functions here


# --------------------- CLASSES SECTION ---------------------
# TODO: define classes here


# ---------------------- MAIN SECTION -----------------------
"""
IMPORTS = "{target}from tinyscript import *\n"
MAIN = """if __name__ == '__main__':
    # TODO: write new arguments
    # Example:
    # parser.add_argument("--option", help="Description")
    initialize()
    # TODO: use validate(...) if necessary
    # TODO: write logic here{base}
"""

TEMPLATES = ["script", "tool"]
TARGETS = {
    "pybots.HTTPBot": """
    with HTTPBot("...", verbose=True) as bot:
        bot.get()
""",
    "pybots.JSONBot": """
    with JSONBot("...", verbose=True) as bot:
        bot.get()
""",
    "pybots.TCPBot": """
    with TCPBot("...", 1234, verbose=True) as bot:
        bot.send_receive("...")
"""
}

NAME_REGEX = re.compile(r'^([0-9a-z]+[-_]+)?[0-9a-z]+$', re.I)


def new(template, target=None, name=None):
    """
    Function for creating a template script or tool.
    
    :param template: template to be used ; one of TEMPLATES
    :param target:   type of script/tool to be created
    :param name:     name of the new script/tool
    """
    if template not in TEMPLATES:
        raise ValueError("Template argument must be one of the followings: {}"
                         .format(", ".join(TEMPLATES)))
    if target is not None and target not in TARGETS.keys():
        raise ValueError("Target argument must be one of the followings: {}"
                         .format(TARGETS.keys()))
    name = name or template
    if NAME_REGEX.match(name) is None:
        raise ValueError("Invalid {} name".format(template))
    with open("{}.py".format(name), 'w') as f:
        target_imp = "" if target is None else "from {} import {}\n" \
                                               .format(*target.split('.'))
        main = MAIN.format(base=TARGETS.get(target) or "")
        if template == "script":
            f.write(SHEBANG + IMPORTS.format(target=target_imp) + "\n\n" + main)
        elif template == "tool":
            f.write(SHEBANG + TOOL_METADATA + TOOL_SECTIONS \
                    .format(imports=IMPORTS.format(target=target_imp)) + main)



# -*- coding: UTF-8 -*-
"""Module for defining hotkeys.

"""
from os import getpid, kill
from signal import SIGINT, SIGTERM, SIGUSR1

from .handlers import _hooks
from .helpers.inputs import confirm, hotkeys, hotkeys_enabled
from .preimports import logging


__all__ = ["set_hotkeys"]


HOTKEYS = None


def __confirm_sig(prompt, sig):
    def _wrapper():
        kill(getpid(), SIGUSR1)
        if confirm(prompt):
            kill(getpid(), sig)
        _hooks.state = "RUNNING"
    return _wrapper


BASE_HOTKEYS = {
    'i': __confirm_sig("Do you really want to interrupt ?", SIGINT),
    'l': logging.lastLogRecord,
    'q': __confirm_sig("Do you really want to quit ?", SIGTERM),
}


def set_hotkeys(glob):
    """
    This function registers the hotkeys got from the global scope.
    
    :param glob: main script's global scope dictionary reference
    """
    if hotkeys_enabled:
        k = glob.get('HOTKEYS', HOTKEYS)
        # case 1: no hotkey to be configured
        if k is None:
            return
        # case 2: only the default hotkeys
        elif k == "default":
            return hotkeys(BASE_HOTKEYS)
        # case 3: only user-defined keys
        elif isinstance(k, dict):
            return hotkeys(k)
        # case 4: default hotkeys and user-defined keys mix
        elif isinstance(k, tuple) and len(k) == 2 and "default" in k:
            r = {}
            for hk in k:
                if hk == "default":
                    hk = BASE_HOTKEYS
                for key, actions in hk.items():
                    r[key] = actions
            return hotkeys(r)
        raise ValueError("Invalid HOTKEYS dictionary")



# -*- coding: UTF-8 -*-
"""Subpackage containing type validation functions and argument types.

"""
from .common import *
from .files import *
from .hash import *
from .network import *
from .strings import *

from .common import __features__ as _common
from .files import __features__ as _files
from .hash import __features__ as _hash
from .network import __features__ as _network
from .strings import __features__ as _strings


__all__ = __features__ = _common + _files + _hash + _network + _strings



# -*- coding: UTF-8 -*-
"""Module for making a report from markdown/HTML to PDF or other individual
 elements to various formats.

"""
from .objects import *
from .report import *

from .objects import __features__ as _objects
from .report import __features__ as _report


__features__ = _objects
__all__ = _report + __features__



# -*- coding: UTF-8 -*-
"""Module for enhancing inspect preimport.

"""
import inspect


def getcallermodule():
    """
    Return the module the caller belongs to.
    """
    return inspect.getmodule(inspect.currentframe().f_back)
inspect.getcallermodule = getcallermodule


def getmainframe():
    """
    Return __main__'s frame object.
    """
    return getparentframe(__name__="__main__")
inspect.getmainframe = getmainframe


def getmainglobals():
    """
    Return __main__'s globals.
    """
    return getmainframe().f_globals
inspect.getmainglobals = getmainglobals


def getmainmodule():
    """
    Return __main__'s frame object.
    """
    return inspect.getmodule(getmainframe())
inspect.getmainmodule = getmainmodule


def getparentframe(**kwargs):
    """
    Return the frame object of the first one having kwargs in its globals.
    """
    frame = inspect.stack()[0][0]
    while frame is not None:
        frame = frame.f_back
        if all(frame.f_globals.get(k) == v for k, v in kwargs.items()):
            break
    return frame
inspect.getparentframe = getparentframe



# -*- coding: UTF-8 -*-
"""Common software licenses.

Source: https://help.github.com/en/articles/licensing-a-repository
"""
from datetime import datetime


__all__ = __features__ = ["copyright", "license", "list_licenses"]


LICENSES = {
    'afl-3.0': "Academic Free License v3.0",
    'agpl-3.0': "GNU Affero General Public License v3.0",
    'apache-2.0': "Apache license 2.0",
    'artistic-2.0': "Artistic license 2.0",
    'bsd-2-clause': "BSD 2-clause \"Simplified\" license",
    'bsd-3-clause': "BSD 3-clause \"New\" or \"Revised\" license",
    'bsd-3-clause-clear': "BSD 3-clause Clear license",
    'bsl-1.0': "Boost Software License 1.0",
    'cc': "Creative Commons license family",
    'cc-by-4.0': "Creative Commons Attribution 4.0",
    'cc-by-sa-4.0': "Creative Commons Attribution Share Alike 4.0",
    'cc0-1.0': "Creative Commons Zero v1.0 Universal",
    'ecl-2.0': "Educational Community License v2.0",
    'epl-1.0': "Eclipse Public License 1.0",
    'eupl-1.1': "European Union Public License 1.1",
    'gpl': "GNU General Public License family",
    'gpl-2.0': "GNU General Public License v2.0",
    'gpl-3.0': "GNU General Public License v3.0",
    'isc': "ISC",
    'lgpl': "GNU Lesser General Public License family",
    'lgpl-2.1': "GNU Lesser General Public License v2.1",
    'lgpl-3.0': "GNU Lesser General Public License v3.0",
    'lppl-1.3c': "LaTeX Project Public License v1.3c",
    'mit': "MIT",
    'mpl-2.0': "Mozilla Public License 2.0",
    'ms-pl': "Microsoft Public License",
    'ncsa': "University of Illinois/NCSA Open Source License",
    'ofl-1.1': "SIL Open Font License 1.1",
    'osl-3.0': "Open Software License 3.0",
    'postgresql': "PostgreSQL License",
    'unlicense': "The Unlicense",
    'wtfpl': "Do What The F*ck You Want To Public License",
    'zlib': "zLib License",
}


def copyright(text, start=None):
    """ Make the copyright field value from the given text. """
    year = datetime.now().year
    if isinstance(start, int) and start < year:
        year = "{}-{}".format(start, year)
    return "© {} {}".format(year, text) if not text.startswith("©") else text


def license(name, null=False):
    """ Get the full license description matching the given short name. """
    return LICENSES.get(str(name)) or ["Invalid license", None][null]


def list_licenses():
    """ Get the list of all supported licenses by short name. """
    return list(LICENSES.keys())




# -*- coding: UTF-8 -*-
"""Module for enhancing code preimport.

"""
import code
import patchy
import types
from collections import deque
from functools import wraps
from six import string_types


BLOCK_KW = ["class", "def", "elif", "else", "except", "finally", "for", "if",
            "try", "while", "with"]
N_MODIF  = 3

is_function  = lambda f: isinstance(f, types.FunctionType)

code.patch   = patchy.patch
code.unpatch = patchy.unpatch
code.Patch   = patchy.temp_patch 


__orig_code = {}
__old_code = {}


class PatchError(ValueError):
    pass
code.PatchError = PatchError


def __apply_code(func, old_code, new_code):
    """ Apply the new code in the given function object. """
    new_code = "\n".join(new_code)
    try:
        patchy.api._set_source(func, new_code)
    except (IndentationError, SyntaxError) as e:
        raise PatchError(e)
    return old_code != new_code


def __get_indent_chars(code_lines):
    """ Guess the indentation characters used in the given code. """
    prev, new = 0, 0
    n, l2 = len(code_lines) - 1, ""
    while prev <= new and n >= 0:
        l1, l2 = l2, code_lines[n]
        indent = l2[:len(l2)-len(l2.lstrip())]
        prev = new
        new = len(indent) if len(l2.strip()) == 0 or \
                             not l2.lstrip().startswith("#") else prev
        n -= 1
    return l1[new:prev]
    

def __get_block_indent(code_lines, n):
    """ Guess the indentation level used in the given code from line n. """
    n %= len(code_lines)
    code_lines[n]  # this will raise an error if n too high
    n -= 1
    l = code_lines[n].split("#", 1)[0].rstrip()  # remove comment
    # find the first previous line which is not empty
    while l.strip() == "" and n >= 0:
        n -= 1
        l = code_lines[n].split("#", 1)[0].rstrip()  # remove comment
    indent = l[:len(l) - len(l.lstrip())]
    # if a block keyword is present, return block's indentation ; otherwise,
    #  return the same indentation
    return indent + __get_indent_chars(code_lines) \
        if any(l.lstrip().startswith(kw) for kw in BLOCK_KW) else indent


def __sort_int_text_pairs(text, lst, item):
    """ Check that l only consists of flattened (int, str or None) pairs and 
         that integers are within the range of the input text. Then output the
         list of sorted pairs. """
    l = len(text.split("\n"))
    s = "Bad code {}s".format(item)
    if len(lst) % 2 != 0:
        raise PatchError(s)
    # check the input list based on multiple criteria
    d = {}
    for i in range(0, len(lst), 2):
        n = lst[i]
        # check for inconsistent index
        if not isinstance(n, int):
            raise PatchError(s + " (bad index {})".format(n))
        n = lst[i] if lst[i] >= 0 else lst[i] + l
        # check for out-of-bound values
        if not 0 <= n < l:
            raise PatchError(s + " (invalid index {})".format(n))
        # check for doubles
        if d.get(n) is not None:
            raise PatchError(s + " (double index {})".format(n))
        # check for inconsistent line
        line = lst[i+1]
        _ = [line, ""][line is None]
        if line is not None and not isinstance(_, string_types):
            raise PatchError(s + " (bad line '{}')".format(line))
        d[n] = line
    # then return the sorted list of pairs
    return sorted(d.items(), reverse=True)


def __validate(f):
    """ Simple validation for validating that the input is a function. """
    if not is_function(f):
        raise ValueError("{} is not a function".format(f))


def _cache(f):
    """ Decorator for caching code changes locally inside the module. """
    @wraps(f)
    def _wrapper(*args, **kwargs):
        cache = kwargs.pop("cache", True)
        if cache:
            func = args[0]
            old_code = code_source(func)
        f(*args, **kwargs)
        if cache:
            if __orig_code.get(func) is None:
                __orig_code[func] = old_code
            if __old_code.get(func) is None:
                __old_code[func] = deque([], N_MODIF)
            __old_code[func].append(old_code)
    return _wrapper


def code_add_block(func, index, block, after=False):
    """
    Additional modification function to allow adding a block of code at a
     specific place in the code of a function. The input block is automatically
     reindented relatively to its insertion position. Therefore, successive
     distinct blocks of code with dedented levels SHALL NOT be added in the same
     block value through this function but well separately.
    
    :param func:  function object to be modified
    :param index: position where the block of code is to be added
    :param block: block of code indented relatively to the insertion position
    :param after: add line after the given index, not before
    :return:      whether code was modified or not
    """
    __validate(func)
    # make the new code
    old_code = code_source(func)
    new_code = old_code.splitlines()
    if after:
        index += 1
    indent = __get_block_indent(new_code, index)
    for line in block.splitlines()[::-1]:
        new_code.insert(index, indent + line)
    return __apply_code(func, old_code, new_code)
code.add_block = code.insert_block = code_add_block


def code_add_line(func, index, addition, **kwargs):
    """
    Alias for applying a single-line addition.
    """
    __validate(func)
    return code_add_lines(func, index, addition, **kwargs)
code.add_line = code.insert_line = code_add_line


@_cache
def code_add_lines(func, *additions, **kwargs):
    """
    Additional modification function to allow adding lines at specific places in
     the code of a function.
    
    :param func:      function object to be modified
    :param additions: list of additions - pairs (line_index, new_line)
    :param after:     add line after the given index, not before
    :return:          whether code was modified or not
    """
    __validate(func)
    after = kwargs.pop("after", False)
    old_code = code_source(func)
    new_code = old_code.splitlines()
    for n, a in __sort_int_text_pairs(old_code, additions, "addition"):
        if after:
            n += 1
        if n == 0:
            new_code.insert(0, a.lstrip())
        else:
            # automatically compute indentation if not provided in added line
            if len(a) == len(a.lstrip()):
                a = __get_block_indent(new_code, n) + a
            new_code.insert(n, a)
    return __apply_code(func, old_code, new_code)
code.add_lines = code.insert_lines = code_add_lines


def code_delete_line(func, index):
    """
    Alias for applying a single-line removal.
    """
    __validate(func)
    return code_delete_lines(func, index)
code.delete_line = code.remove_line = code_delete_line


def code_delete_lines(func, *indices):
    """
    Additional function to allow deleting only specific lines in the code of a
     function.
    
    :param func:    function object to be modified
    :param indices: list of line indices for removal
    :return:        whether code was modified or not
    """
    __validate(func)
    replacements = []
    for i in indices:
        replacements.extend([i, None])
    return code_replace_lines(func, *replacements)
code.delete_lines = code.remove_lines = code_delete_lines


@_cache
def code_replace(func, *replacements):
    """
    Slight modification to original replace function to allow replacing only
     part(s) of code and not necessarily using whole function's code.
    
    :param func:         function object to be modified
    :param replacements: list of replacements - pairs (str_to_repl, replac)
    :return:             whether code was modified or not
    """
    __validate(func)
    try:
        patchy.replace(func, *replacements)
        old_code, new_code = replacements
    # if normal handling fails, get whole function's code first, apply
    #  replacement then retry
    except:
        if len(replacements) % 2 != 0:
            raise PatchError("Bad code replacement")
        old_code = code_source(func)
        new_code = old_code
        for i in range(0, len(replacements), 2):
            new_code = new_code.replace(replacements[i], replacements[i+1])
        patchy.replace(func, old_code, new_code)
    # report whether the code was changed or not
    return old_code != new_code
code.replace = code_replace


def code_replace_line(func, index, replacement):
    """
    Alias for applying a single-line replacement.
    """
    __validate(func)
    return code_replace_lines(func, index, replacement)
code.replace_line = code_replace_line


@_cache
def code_replace_lines(func, *replacements):
    """
    Additional replace function to allow replacing only specific lines in the
     code of a function.
    
    :param func:         function object to be modified
    :param replacements: list of replacements - pairs (line_index, replac)
    :return:             whether code was modified or not
    """
    __validate(func)
    r = replacements
    old_code = code_source(func)
    new_code = old_code.splitlines()
    for n, r in __sort_int_text_pairs(old_code, replacements, "replacement"):
        if r is None:
            new_code.pop(n)
        else:
            l = new_code[n]
            indent = l[:len(l)-len(l.lstrip())]
            new_code[n] = indent + r.lstrip()
    return __apply_code(func, old_code, new_code)
code.replace_lines = code_replace_lines


def code_restore(func):
    """
    Additional function to restore mofidied code.
    
    :param func: function object to be restored
    :return:     True if function's code was restored
    """
    __validate(func)
    old_code = __orig_code.get(func)
    if old_code is not None:
        patchy.api._set_source(func, patchy.api.dedent(old_code))
        __old_code[func] = deque([], N_MODIF)
        return True
    return False
code.restore = code_restore


def code_revert(func):
    """ 
    Additional function to revert mofidied code to last modification.
    
    :param func: function object to be reverted to last version
    :return:     True if function's code was reverted to last version
    """
    __validate(func)
    try:
        old_code = __old_code.get(func).pop()
        if old_code is not None:
            patchy.api._set_source(func, patchy.api.dedent(old_code))
            return True
    except:
        return False
code.revert = code_revert


def code_source(func):
    """ 
    Alias function to get function's code.
    
    :param func: function object
    :return:     function's source code
    """
    __validate(func)
    return patchy.api._get_source(func)
code.source = code_source


# disable source AST check to avoid Python compatibility errors
code_replace(patchy.api._assert_ast_equal,
             "current_ast = ast.parse",
             "return #current_ast = ast.parse",
             cache=False)



# -*- coding: UTF-8 -*-
"""Module for defining the list of preimports.

"""
import codext
from importlib import import_module
try:  # will work in Python 3
    from importlib import reload
    import configparser
except ImportError:  # will fail in Python 2 ; it will keep the built-in reload
    reload = reload
    import ConfigParser as configparser

from .codep import code
from .hash import hashlib
from .inspectp import inspect
from .log import logging
from .rand import random
from .venv import virtualenv, PipPackage, VirtualEnv


__all__ = __features__ = ["PipPackage", "VirtualEnv", "import_module"]
__all__ += ["__imports__", "load", "reload"]

__imports__ = {
    'bad': [],
    'enhanced': [
        "code",
        "hashlib",
        "inspect",
        "logging",
        "random",
        "virtualenv",
    ],
    'standard': [
        "argparse",
        "ast",
        "base64",
        "binascii",
        "codecs",
        "collections",
        "configparser",
        "ctypes",
        "fileinput",
        "itertools",
        "os",
        "platform",
        "re",
        "shutil",
        "signal",
        "string",
        "struct",
        "subprocess",
        "sys",
        "time",
        "types",
    ],
    'optional': [
        "bs4",
        "fs",
        "loremipsum",
        "requests",
    ]
}


def _load_preimports(*extras):
    """
    This loads the list of modules to be preimported in the global scope.
    
    :param extra: additional modules
    :return:      list of successfully imported modules, list of failures
    """
    i = __imports__
    for module in i['standard'] + i['enhanced'] + list(extras):
        load(module)
    for module in i['optional']:
        load(module, True)


def load(module, optional=False):
    """
    This loads a module and, in case of failure, appends it to a list of bad
     imports or not if it is required or optional.
    
    :param module:   module name
    :param optional: whether the module is optional or not
    """
    global __features__, __imports__
    m = globals().get(module)
    if m is not None:  # already imported (e.g. configparser)
        __features__.append(module)
        return m
    try:
        globals()[module] = m = import_module(module)
        m.__name__ = module
        __features__.append(module)
        return m
    except ImportError:
        if not optional and module not in __imports__['bad']:
            __imports__['bad'].append(module)
            for k, l in __imports__.items():
                if k != 'bad' and module in l:
                    l.remove(module)


_load_preimports()



# -*- coding: UTF-8 -*-
import ctypes
import os
from importlib import import_module
from pathlib import Path as BasePath
from random import choice
from re import search
from shutil import rmtree
from six import string_types
from tempfile import gettempdir, NamedTemporaryFile as TempFile

from .constants import *
from .compat import u


__all__ = __features__ = ["Path", "MirrorPath", "TempPath"]


class Path(BasePath):
    """ Extension of the base class Path from pathlib.
    
    :param expand: expand user's path
    :param create: create the directory if it doesn't exist
    """
    _flavour = BasePath()._flavour  # fix to AttributeError
    
    def __new__(cls, *parts, **kwargs):
        expand = kwargs.pop("expand", False)
        create = kwargs.pop("create", False)
        _ = super(Path, cls).__new__(cls, *parts, **kwargs)
        if expand:
            _ = _.expanduser().absolute()
        if create and not _.exists():
            _.mkdir(parents=True)  # exist_ok does not work in Python 2
        return _
    
    @property
    def basename(self):
        """ Dummy alias for name attribute. """
        return self.name
    
    @property
    def bytes(self):
        """ Get file's content as bytes. """
        with self.open('rb') as f:
            return f.read()
    
    @property
    def child(self):
        """ Get the child path relative to self's one. """
        return Path(*self.parts[1:])
    
    @property
    def filename(self):
        """ Get the file name, without the complete path. """
        return self.stem + self.suffix
    
    @property
    def size(self):
        """ Get path's size. """
        if self.is_file() or self.is_symlink():
            return self.stat().st_size
        elif self.is_dir():
            s = 4096  # include the size of the directory itself
            for root, dirs, files in os.walk(str(self)):
                s += 4096 * len(dirs)
                for f in files:
                    s += os.stat(str(Path(root).joinpath(f))).st_size
            return s
    
    @property
    def text(self):
        """ Get file's content as a string. """
        return self.read_text()
    
    def __add_text(self, data, mode='w', encoding=None, errors=None):
        """ Allows to write/append text to the file, both in Python 2 and 3. """
        if not isinstance(data, string_types):
            raise TypeError("data must be str, not %s" % 
                            data.__class__.__name__)
        with self.open(mode=mode, encoding=encoding, errors=errors) as f:
            return f.write(u(data))
    
    def append_bytes(self, data):
        """ Allows to append bytes to the file, as only write_bytes is available in pathlib, overwritting the former
             bytes at each write. """
        with self.open(mode='ab') as f:
            return f.write(memoryview(data))
    
    def append_line(self, line):
        """ Shortcut for appending a single line (text with newline). """
        self.append_text(["\n", ""][self.size == 0] + line)
    
    def append_lines(self, *lines):
        """ Shortcut for appending a bunch of lines. """
        for line in lines:
            self.append_line(line)
    
    def append_text(self, text, encoding=None, errors=None):
        """ Allows to append text to the file, as only write_text is available in pathlib, overwritting the former text
             at each write. """
        return self.__add_text(text, 'a', encoding, errors)
    
    def choice(self, *filetypes):
        """ Return a random file from the current directory. """
        if not self.is_dir():
            return self
        filetypes = list(filetypes)
        while len(filetypes) > 0:
            filetype = choice(filetypes)
            filetypes.remove(filetype)
            l = list(self.iterfiles(filetype, filename_only=True))
            if len(l) > 0:
                return self.joinpath(choice(l))
    
    def expanduser(self):
        """ Fixed expanduser() method, working for both Python 2 and 3. """
        return Path(os.path.expanduser(str(self)))
    
    def find(self, name=None, regex=False):
        """ Find a folder or file from the current path. """
        if name is None:
            f = lambda p: True
        elif not regex:
            if "*" not in name:
                f = lambda p: p.basename == name
            else:
                r = "^{}$".format(name.replace(".", "\\.").replace("?", "\\?").replace("-", "\\-").replace("+", "\\+")
                                      .replace("[", "\\[").replace("]", "\\]").replace("(", "\\(").replace(")", "\\)")
                                      .replace("{", "\\{").replace("}", "\\}").replace("*", ".*"))
                f = lambda p: search(r, p.basename) is not None
        else:
            f = lambda p: search(name, p.basename) is not None
        for item in self.walk(filter_func=f):
            yield item
    
    def generate(self, prefix="", suffix="", length=8, alphabet="0123456789abcdef"):
        """ Generate a random folder name. """
        if not self.is_dir():
            return self
        while True:
            _ = "".join(choice(alphabet) for i in range(length))
            new = self.joinpath(str(prefix) + _ + str(suffix))
            if not new.exists():
                return new
    rand_folder_name = generate
    
    def is_hidden(self):
        """ Check if the current path is hidden. """
        if DARWIN:
            fnd = import_module("Foundation")
            u, f = fnd.NSURL.fileURLWithPath_(str(self)), fnd.NSURLIsHiddenKey
            return u.getResourceValue_forKey_error_(None, f, None)[1]
        elif LINUX:
            return self.stem.startswith(".")
        elif WINDOWS:
            import win32api, win32con
            return win32api.GetFileAttributes(p) & \
                   (win32con.FILE_ATTRIBUTE_HIDDEN | \
                    win32con.FILE_ATTRIBUTE_SYSTEM)
        raise NotImplementedError("Cannot check for the hidden status on this platform")
    
    def is_samepath(self, otherpath):
        """ Check if both paths have the same parts. """
        return self.absolute().parts == Path(otherpath).absolute().parts
    
    def iterfiles(self, filetype=None, filename_only=False):
        """ List all files from the current directory. """
        for i in self.iterdir():
            if i.is_file() and (filetype is None or i.suffix == filetype):
                yield i.filename if filename_only else i
    
    def iterpubdir(self):
        """ List all visible subdirectories from the current directory. """
        for i in self.iterdir():
            if i.is_dir() and not i.is_hidden():
                yield i
    
    def listdir(self, filter_func=lambda p: True, sort=True):
        """ List the current path using the given filter. """
        l = os.listdir(str(self))
        if sort:
            l = sorted(l)
        for item in l:
            item = self.joinpath(item)
            if filter_func(item):
                yield item
    
    def mkdir(self, mode=0o777, parents=False, exist_ok=False):
        """ Fix to non-existing argument exist_ok in Python 2. """
        arg = (exist_ok, ) if PYTHON3 else ()
        super(Path, self).mkdir(mode, parents, *arg)
    
    def read_lines(self, encoding=None, errors=None):
        """ Extra method for reading a file as lines. """
        for l in self.read_text(encoding, errors).splitlines():
            yield l
    
    def read_text(self, encoding=None, errors=None):
        """ Fix to non-existing method in Python 2. """
        with self.open(mode='r', encoding=encoding, errors=errors) as f:
            return f.read()
    
    def reset(self):
        """ Ensure the file exists and is empty. """
        with self.open('w') as f:
            pass
    
    def remove(self):
        """ Extension for removing a directory or a file. """
        if self.is_dir():
            rmtree(str(self))
        else:
            os.remove(str(self))
    
    def walk(self, breadthfirst=True, filter_func=lambda p: True, sort=True):
        """ Walk the current path for directories and files using os.listdir(), breadth-first or depth-first, sorted or
             not, based on a filter function. """
        if breadthfirst:
            for item in self.listdir(lambda p: not p.is_dir(), sort):
                if filter_func(item):
                    yield item
        for item in self.listdir(lambda p: p.is_dir(), sort):
            if breadthfirst and filter_func(item):
                yield item
            for subitem in item.walk(breadthfirst, filter_func):
                yield subitem
            if not breadthfirst and filter_func(item):
                yield item
        if not breadthfirst:
            for item in self.listdir(lambda p: not p.is_dir(), sort):
                if filter_func(item):
                    yield item
    
    def write_text(self, data, encoding=None, errors=None):
        """ Fix to non-existing method in Python 2. """
        return self.__add_text(data, 'w', encoding, errors)


class MirrorPath(Path):
    """ Extension of the class Path for handling a folder that can mirror another one using symbolic links.
    
    :param destination: destination folder where the structure is mirrored
    :param source:      source folder from which the structure is mirrored
    """
    def __new__(cls, destination, source=None, **kwargs):
        self = super(MirrorPath, cls).__new__(cls, destination, **kwargs)
        if source is not None:
            self.mirror(source)
        return self
    
    def mirror(self, source):
        """ Mirror the source item. """
        self._mirrored = []
        source = Path(source)
        if not self.exists():
            self.symlink_to(str(source.absolute()), source.is_dir())
        elif not self.is_symlink() and source.is_dir():
            for item in os.listdir(str(source)):
                dst, src = self.joinpath(item), source.joinpath(item)
                self._mirrored.append(MirrorPath(dst, src))
    
    def unmirror(self):
        """ Perform the reverse operation of mirror. """
        while len(self._mirrored) > 0:
            self._mirrored.pop(0).unmirror()
        if self.is_symlink():
            self.unlink()


class TempPath(Path):
    """ Extension of the class Path for handling a temporary path.
    
    :param length:   length for the folder name (if 0, do not generate a folder name, e.g. keeping /tmp)
    :param alphabet: character set to be used for generating the folder name
    """
    def __new__(cls, **kwargs):
        kw = {}
        kw["prefix"]   = kwargs.pop("prefix", "")
        kw["suffix"]   = kwargs.pop("suffix", "")
        kw["length"]   = kwargs.pop("length", 0)
        kw["alphabet"] = kwargs.pop("alphabet", "0123456789abcdef")
        _ = Path(gettempdir())
        kwargs["create"] = True   # force creation
        kwargs["expand"] = False  # expansion is not necessary
        if kw["length"] > 0:
            while True:
                # ensure this is a newly generated path
                tmp = _.generate(**kw)
                if not tmp.exists():
                    break
            return super(TempPath, cls).__new__(cls, tmp, **kwargs)
        return super(TempPath, cls).__new__(cls, _, **kwargs)
    
    def tempfile(self, **kwargs):
        """ Create a NamedTemporaryFile in the TempPath. """
        kwargs.pop("dir", None)
        return TempFile(dir=str(self), **kwargs)




# -*- coding: UTF-8 -*-
"""Class property.

"""
__all__ = __features__ = ["classproperty"]


def classproperty(f):
    if not isinstance(f, (classmethod, staticmethod)):
        f = classmethod(f)
    return ClassProperty(f)


class ClassProperty(object):
    """
    ClassProperty class for implementing @classproperty.
    
    Inspired from: https://stackoverflow.com/questions/5189699/how-to-make-a-class-property
    """
    def __init__(self, fget=None, fset=None, doc=None):
        self.fget = fget
        self.fset = fset
        self.__doc__ = doc or getattr(fget, "__doc__", None)
    
    def __get__(self, obj, objtype=None):
        objtype = objtype or type(obj)
        return self.fget.__get__(obj, objtype)()
    
    def setter(self, f):
        if not isinstance(f, (classmethod, staticmethod)):
            f = classmethod(f)
        self.fset = f
        return self




# -*- coding: UTF-8 -*-
"""Report data transformation functions.

"""
import re
import xmltodict
from collections import OrderedDict
from dicttoxml import dicttoxml
from inspect import currentframe
from json2html import json2html as j2h


__all__ = __features__ = ["dict2html", "dict2xml", "json2html", "json2xml", "report2objects", "xml2dict", "xml2json"]


json2html = dict2html = j2h.convert
json2xml = dict2xml = dicttoxml
xml2json = xml2dict = xmltodict.parse


def report2objects(text, header_sep=None, footer_sep=None):
    """ Convert a raw text report (i.e. WPScan-like) to Tinyscript report objects. """
    glob = currentframe().f_back.f_globals
    o = glob.get('Report', list)()
    if header_sep:
        parts = re.split("[" + re.escape(header_sep) + "]{10,}", text)
        if len(parts) > 1:
            header = parts.pop(0).strip()
            while len(parts) > 0 and header == "":
                header = parts.pop(0).strip()
            if header != "":
                Header = glob.get('Header', None)
                o.append(("Header", header) if Header is None else Header(header))
                text = parts[0]
    if footer_sep:
        parts = re.split("[" + re.escape(footer_sep) + "]{10,}", text)
        if len(parts) > 1:
            footer = parts.pop().strip()
            while len(parts) > 0 and footer == "":
                footer = parts.pop().strip()
            if footer != "":
                Footer = glob.get('Footer', None)
                o.append(("Footer", footer) if Footer is None else Footer(footer))
                text = parts[0]
    blocks = list(re.split(r"(?:\r?\n){2,}", text))
    for i, block in enumerate(blocks):
        block = block.strip()
        lines = re.split(r"\r?\n", block)
        if len(lines) == 1:
            if re.match(r"\[.\]\s", block):
                Subsection = glob.get('Subsection', None)
                o.append(("Subsection", block[4:]) if Subsection is None else Subsection(block[4:]))
            else:
                Section = glob.get('Section', None)
                o.append(("Section", block) if Section is None else Section(block))
        else:
            Text = glob.get('Text', None)
            o.append(("Text", block) if Text is None else Text(block))
    return o




# -*- coding: UTF-8 -*-
"""Common utility functions.

"""
import colorful
import os
import re
import signal
import sys
from ast import literal_eval
from getpass import getuser
from six import StringIO

from .compat import ensure_str
from .constants import *
from .data.types import is_function, is_lambda, is_str

# fix to Xlib.error.DisplayConnectionError: Can't connect to display ":0": No protocol specified
# however, it does not fix the error while testing with Travis CI
if LINUX:
    os.system("xhost +SI:localuser:{} > /dev/null 2>&1".format(getuser()))
os.environ['DISPLAY'] = os.environ.get('DISPLAY') or os.environ.get('REMOTE_DISPLAY', ":0")

try:
    from pynput.keyboard import Controller, Key, Listener
    _keyboard = Controller()
    hotkeys_enabled = True
except Exception:  # catch ImportError but also Xlib.error.DisplayConnectionError
    _keyboard = None
    hotkeys_enabled = False


__all__ = __features__ = ["capture", "clear", "confirm", "hotkeys", "pause", "silent", "std_input", "stdin_flush",
                          "stdin_pipe", "user_input", "Capture"]

pause = lambda *a, **kw: std_input("Press Enter to continue", *a, **kw) or None


def capture(f):
    """
    Decorator for capturing stdout and stderr.
    """
    def _wrapper(*a, **kw):
        with Capture() as (out, err):
            r = f(*a, **kw)
        return r, out.text, err.text
    return _wrapper


def clear():
    """
    Dummy multi-platform screen clear function.
    """
    from os import system
    if DARWIN or LINUX:
        system("clear")
    elif WINDOWS:
        system("cls")


def confirm(prompt="Are you sure ?", style="bold"):
    """
    Ask for confirmation.
    """
    return user_input("\r" + prompt, ["(Y)es", "(N)o"], "n", style=style) == "yes"


def hotkeys(hotkeys, silent=True):
    """
    Hotkeys declaration function relying on pynput.
    
    :param hotkeys: dictionary of hotkeys and related actions
    :param silent:  do not show errors (of keys not handled)
    :param die:     whether the hotkey should have its listener respawn or not
    
    Hotkeys dictionary formats:
    1. hotkey: on_press action
    2. hotkey: dictionary with on_press and on_release actions
    
    Action representations:
    1. (str, output handler)
    2. function returning None|str|(str, output handler)
    """
    if not hotkeys_enabled:
        return
    global listener, on_press
    # close the running listener, if hotkeys(...) was already called
    try:
        listener.stop()
        del listener
    except NameError:
        pass
    # replace string keys to Key.[...] objects (e.g. 'ctrl')
    for k, v in hotkeys.items():
        try:
            k_obj = getattr(Key, k)
            hotkeys[k_obj] = v
            del hotkeys[k]
        except Exception:
            pass
    
    def __handle(f):
        """ Hotkey action handler and listener regenerator. """
        global listener, on_press
        # stop the listener to avoid capturing multiple keystrokes
        listener.stop()
        # flush stdin to avoid printing the keystrokes
        stdin_flush()
        r = f()
        r, out = r if isinstance(r, tuple) and len(r) == 2 else (r, None)
        if is_str(r):
            if out is None:
                print(r)
            elif out in [sys.stdout, sys.stderr]:
                out.write(r)
                out.flush()
            else:
                out(r)
        # open a new listener
        listener = Listener(on_press=on_press)
        listener.start()
    
    def on_press(key):
        """ Generic on_press callback function. """
        try:
            k = key._name_
        except AttributeError:
            k = key.char
        try:
            v = hotkeys[ensure_str(k)]
        except KeyError:
            if not silent:
                raise ValueError("Key '{}' not handled".format(k))
            return
        __handle((lambda: v) if not is_function(v) else v)
    
    listener = Listener(on_press=on_press)
    listener.start()


def silent(f):
    """
    Decorator for silencing stdout and stderr.
    """
    def _wrapper(*a, **kw):
        with Capture():
            r = f(*a, **kw)
        return r
    return _wrapper


def std_input(prompt="", style=None, palette=None):
    """
    Very simple Python2/3-compatible input function handling prompt styling.
    
    :param prompt:  prompt message
    :param style:   colorful styling function, e.g. red_on_green (for green foreground and red background colors)
    :param palette: dictionary for defining new styles
    """
    colorful.update_palette(palette or {})
    if style is not None:
        if isinstance(style, (list, tuple, set)):
            style = "_".join(style)
        prompt = getattr(colorful, style)(prompt)
    return (input(prompt) if PYTHON3 else raw_input(prompt)).strip()


def stdin_flush():
    """
    Multi-platform stdin flush function.
    
    Source: https://rosettacode.org/wiki/Keyboard_input/Flush_the_keyboard_buffer#Python
    """
    try:
        try:  # Windows
            from msvcrt import getch, kbhit
            while kbhit(): getch()
        except ImportError:  # Linux/Unix
            from termios import tcflush, TCIOFLUSH
            tcflush(sys.stdin, TCIOFLUSH)
    except Exception:
        pass


def stdin_pipe():
    """
    Python2/3-compatible stdin pipe read function.
    """
    if PYTHON3:
        with open(0, 'rb') as f:
            for l in f:
                yield l
    else:
        for l in sys.stdin:
            yield l


def user_input(prompt="", choices=None, default=None, choices_str="", required=False, newline=False, **kwargs):
    """
    Python2/3-compatible input function handling choices and default value.
    
    :param prompt:      prompt message
    :param choices:     list of possible choices or lambda function
    :param default:     default value
    :param choices_str: list of possible choices as a string (overrides the default composition from the choices list)
    :param required:    make non-null user input mandatory
    :param newline:     insert a newline and '>>' after the prompt
    :param kwargs:      keyword-arguments to be passed to std_input for styling
    :return:            handled user input
    """
    shortcuts = {}
    if type(choices) in [list, tuple, set]:
        choices = list(map(lambda x: str(x).lower(), choices))
        choices_str = " {%s}" % (choices_str or '|'.join(list(map(str, choices))))
        # consider choices of the form ["(Y)es", "(N)o"] ; in this case, we want the choices to be
        #  ['yes', 'no', 'y', 'n'] for the sake of simplicity for the user
        m = list(map(lambda x: re.match(r'\(([a-zA-Z0-9])\)', x), choices))
        # then remove the parenthesis from the choices
        choices = [re.sub(r"\(([a-zA-Z0-9])\)", lambda x: x.group(1).lower(), c) for c in choices]
        shortcuts = {x.group(1).lower(): c for x, c in zip(m, choices) if x is not None}
        # this way, if using ["Yes", "No"], choices will remain so
        _check = lambda v: str(v).lower() in choices or str(v).lower() in shortcuts.keys()
    elif is_lambda(choices):
        _check = choices
    else:
        _check = lambda v: True
    prompt += "{}{} ".format(choices_str, [" [{}]".format(default), ""][default is None and required])
    user_input, first = None, True
    while not user_input:
        stdin_flush()
        user_input = std_input(["", prompt][first] + ["", "\n >> "][newline], **kwargs)
        first = False
        if type(choices) in [list, tuple, set]:
            choices = list(map(lambda x: x.lower(), choices))
            user_input = user_input.lower()
        if user_input == "" and default is not None and _check(default):
            _ = str(default)
            return shortcuts.get(_) or _
        if user_input != "" and _check(user_input):
            _ = user_input
            return shortcuts.get(_) or _
        if not required:
            return


class _Text(object):
    """
    Dummy Text class for storing StringIO's content before closing it.
    """
    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.text


class Capture(object):
    """
    Context manager for capturing stdout and stderr.
    """
    def __init__(self, out=sys.stdout, err=sys.stderr):
        # backup original output file handles
        self._stdout = sys.stdout
        self._stderr = sys.stderr
        
    def __enter__(self):
        # create new file handles
        sys.stdout, sys.stderr = StringIO(), StringIO()
        self.stdout, self.stderr = _Text(), _Text()
        # return references of the dummy objects
        return self.stdout, self.stderr
    
    def __exit__(self, *args):
        # freeze stdout and stderr contents before closing the file handles, using the references set in __enter__
        self.stdout.text = sys.stdout.getvalue().strip() 
        self.stderr.text = sys.stderr.getvalue().strip()
        # close current file handles
        sys.stdout.close()
        sys.stderr.close()
        # restore original output file handles
        sys.stdout, sys.stderr = self._stdout, self._stderr




# -*- coding: UTF-8 -*-
"""Common utility functions for data.

"""
import binascii
import bitstring
import patchy
import re
from ast import literal_eval as litev
from random import randint

from .types import *


__all__ = __features__ = ["BitArray", "pad", "unpad"]


PAD = ["ansic9.23", "incremental", "iso7816-4", "pkcs5", "pkcs7", "w3c"]


patchy.replace(bitstring.Bits._getlength, """
    def _getlength(self):
        \"\"\"Return the length of the bitstring in bits.\"\"\"
        return self._datastore.bitlength
""", """
    def _getlength(self):
        \"\"\"Return the length of the bitstring in bits.\"\"\"
        l = self._datastore.bitlength
        return l + (8 - l % 8) % 8 if getattr(Bits, "_padding", True) else l
""")
patchy.replace(bitstring.Bits._getbin, """
    def _getbin(self):
        \"\"\"Return interpretation as a binary string.\"\"\"
        return self._readbin(self.len, 0)
""", """
    def _getbin(self):
        \"\"\"Return interpretation as a binary string.\"\"\"
        Bits._padding = False
        r = self._readbin(self.len, 0)
        Bits._padding = True
        return r
""")


class BitArray(bitstring.BitArray):
    __doc__ = """
    Small improvement to the original bitstring.BitArray class.
    
    It allows to set the number of bits per group (by default, 8, for considering bytes).
    
    """ + bitstring.BitArray.__doc__
    
    def __new__(cls, auto=None, length=None, offset=None, nbits=8, **kwargs):
        if auto:
            if not auto.startswith("0b"):
                auto = "0b" + auto
        c = super(BitArray, cls).__new__(cls, auto, length, offset, **kwargs)
        c._nbits = nbits
        c.original = True
        return c
    
    @property
    def nbits(self):
        return self._nbits
    
    @nbits.setter
    def nbits(self, n):
        ob = self.bin
        nb = ""
        for i in range(0, len(ob), self._nbits):
            group = ob[i:i+self._nbits]
            if i > 0 and len(group) < self._nbits:
                break
            group = pad(group, ">0", n)[-n:]
            if int(ob[i:i+self._nbits], 2) != int(group, 2):
                self.original = False
            nb += group
        self.bin = nb
        self._nbits = n


def pad(string, padding=None, blocksize=8, raw=False):
    """
    Custom string padding function relying on Tinyscript's data type check functions and a few implemented padding
     algorithms.
    
    NB: Default algorithm is the zero padding.
    
    :param padding:   padding characters or padding algorithm
    :param blocksize: block size to be considered for padding
    :param raw:       whether the input string must be considered as a normal string (disables the check for hexstring)
    """
    s, bs = string, blocksize
    isb, ish = is_bin(s) and not raw, is_hex(s) and not raw
    if padding in PAD and isb or padding == "bit" and not isb:
        raise ValueError("Incompatible padding algorithm")
    if blocksize <= 0:
        raise ValueError("Block size must be a positive integer")
    to_char = lambda x: [chr(x), "{:0>2}".format(hex(x)[2:])][ish]
    lp = [1, 2][ish]
    bn = bs * lp
    p = int((bn - (len(s) % bn) / lp) % bs)
    zc = ["\x00", "00"][ish]
    nc = to_char(p)
    if padding == "ansic9.23":
        return s + (p - 1) * zc + int(p > 0) * nc
    elif padding == "bit":
        p = bs - (len(s) % bs) % bs
        return s + "1" + (p - 1) * "0"
    elif padding == "incremental":
        return s + "".join(map(to_char, range(1, p + 1)))
    elif padding == "iso7816-4":
        return s + ["\x80", "80"][ish] + (p - 1) * zc
    elif padding in ["pkcs5", "pkcs7"]:
        return s + p * nc
    elif padding == "w3c":
        rand = "".join(to_char(randint(0, 255)) for i in range(p - 1))
        return s + rand + int(p > 0) * nc
    else:
        padding = padding or ("0" if isb else "00" if ish else "\x00")
        if re.match(r"[<>]", padding):
            left = padding[0] == ">"
            padding = padding[1:]
        else:
            left = False
        if len(s) % len(padding) != 0:
            raise ValueError("Bad padding character(s)")
        lp = len(padding)
        bn = bs * lp
        p = int((bn - (len(s) % bn) / lp) % bs)
        return ["", p * padding][left] + s + [p * padding, ""][left]


def unpad(string, padding=None, blocksize=8, raw=False):
    """
    Custom string unpadding function relying on Tinyscript's data type check functions and a few implemented padding
     algorithms.
    
    NB: Default algorithm is the zero padding.
    
    :param padding:   padding characters or padding algorithm
    :param blocksize: block size to be considered for padding
    :param raw:       whether the input string must be considered as a normal string (disables the check for hexstring)
    """
    s, bs = string, blocksize
    isb, ish = is_bin(s) and not raw, is_hex(s) and not raw
    if padding in PAD and isb or padding == "bit" and not isb:
        raise ValueError("Incompatible padding algorithm")
    if blocksize <= 0:
        raise ValueError("Block size must be a positive integer")
    to_char = lambda x: [chr(x), "{:0>2}".format(hex(x)[2:])][ish]
    lastb = [litev("0x" + (s[i:i+2 or len(s)] or "0")) for i in range(-bs*2, 0, 2)] if ish else list(map(ord, s[-bs:]))
    n = lastb[-1]
    lp = [1, 2][ish]
    if padding == "ansic9.23":
        if n <= bs and lastb[-n:-1] == [0] * (n - 1):
            return s[:-n*lp]
    elif padding == "bit":
        l = len(s)
        s = s.rstrip("0")
        if len(s) < l:
            return s[:-1]
    elif padding == "incremental":
        if n <= bs and lastb[-n:-1] == list(range(1, n)):
            return s[:-n*lp]
    elif padding == "iso7816-4":
        n = 0
        while lastb[-1] == 0:
            lastb.pop()
            n += 1
        if n < bs and lastb[-1] == 128:
            lastb.pop()
            n += 1
            return s[:-n*lp]
    elif padding in ["pkcs5", "pkcs7"]:
        if n <= bs and lastb[-n:-1] == (n - 1) * [n]:
            return s[:-n*lp]
    elif padding == "w3c":
        n = lastb[-1]
        if n <= bs:
            return s[:-n*lp or len(s)]
    else:
        return s.rstrip("0") if isb else pad(s.rstrip("0"), "0", 2) if ish else s.rstrip("\x00")
    return s




# -*- coding: UTF-8 -*-
"""Hash-related checking functions and argument types.

"""
import re


__all__ = __features__ = []


# hash check functions
__all__ += ["is_hash", "is_md5", "is_sha1", "is_sha224", "is_sha256",
            "is_sha384", "is_sha512"]
is_hash   = lambda h: any(__check_hash(h, a, False) is not None for a in \
                          HASH_LEN.keys())
is_md5    = lambda h: __check_hash(h, "md5", False) is not None
is_sha1   = lambda h: __check_hash(h, "sha1", False) is not None
is_sha224 = lambda h: __check_hash(h, "sha224", False) is not None
is_sha256 = lambda h: __check_hash(h, "sha256", False) is not None
is_sha384 = lambda h: __check_hash(h, "sha384", False) is not None
is_sha512 = lambda h: __check_hash(h, "sha512", False) is not None


# hash-related argument types
__all__ += ["any_hash", "md5_hash", "sha1_hash", "sha224_hash", "sha256_hash",
            "sha512_hash"]
HASH_LEN = {'md5': 32, 'sha1': 40, 'sha224': 56, 'sha256': 64, 'sha384': 96,
            'sha512': 128}


def __check_hash(s, algo, fail=True):
    l = HASH_LEN[algo]
    if re.match(r"(?i)^[a-f0-9]{%d}$" % l, s) is None:
        if fail:
            raise ValueError("Bad {} hash".format(algo))
        return
    return s
md5_hash    = lambda h: __check_hash(h, "md5")
sha1_hash   = lambda h: __check_hash(h, "sha1")
sha224_hash = lambda h: __check_hash(h, "sha224")
sha256_hash = lambda h: __check_hash(h, "sha256")
sha384_hash = lambda h: __check_hash(h, "sha384")
sha512_hash = lambda h: __check_hash(h, "sha512")


def any_hash(h):
    if not any(__check_hash(h, a, False) is not None for a in HASH_LEN.keys()):
        raise ValueError("Bad hash")
    return h



# -*- coding: UTF-8 -*-
"""Module for defining common functions for report elements.

"""
from functools import wraps
from os.path import exists, splitext
from six import string_types


__all__ = ["output", "string_types", "Element", "TEXT"]

TEXT = True


def output(f):
    """ This decorator allows to choose to return an output as text or to save
         it to a file. """
    f._output = True
    @wraps(f)
    def _wrapper(self, *args, **kwargs):
        try:
            text = kwargs.get('text') or args[0]
        except IndexError:
            text = True
        r = f(self, *args, **kwargs)
        if text:
            return r
        elif r is not None and isinstance(r, string_types):
            filename = "{}.{}".format(self.filename, f.__name__)
            while exists(filename):
                name, ext = splitext(filename)
                try:
                    name, i = name.split('-')
                    i = int(i) + 1
                except ValueError:
                    i = 2
                filename = "{}-{}".format(name, i) + ext
            with open(filename, 'w') as out:
                out.write(r)
    return _wrapper


class Element(object):
    """ This class is used to give a common type to report elements. """
    _style = {'size': 12, 'style': "normal", 'color': "black"}
    id = 0
    
    def __init__(self, **kwargs):
        self.name = kwargs.get('name', self.__class__.__name__.lower())
        self.css = ""
        for k, v in kwargs.items():
            if k not in ['color', 'size', 'style']:
                continue
            self._style[k] = v
        self.style = "font-size:%(size)spx;font-style:%(style)s;color:%(color)s;" % self._style
        self._newline = "\n"
    
    def __repr__(self):
        return "<{}: {}>".format(self.__class__.__name__, self.name)
    
    def _set_indent(self, indent):
        return ("", "") if indent is None else (indent * " ", "\n")
    
    @output
    def csv(self, text=TEXT):
        return ""
    
    @output
    def html(self, indent=4, text=TEXT):
        return ""
    
    @output
    def json(self, text=TEXT):
        return {self.name: self.data}
    
    @output
    def md(self, text=TEXT):
        return ""
    
    @output
    def xml(self, indent=2, text=TEXT):
        return ("<%(name)s>{0}{1}{0}</%(name)s>" % self.__dict__).format(self._newline, str(self.data))




# -*- coding: UTF-8 -*-
"""Module for enhancing hashlib preimport.

"""
import hashlib
from os.path import expanduser, isfile
from six import b


def hash_file(filename, algo="sha256"):
    """
    This extends hashlib's hashing function to hash a file per block.

    :param filename: name of the file to be hashed
    :return:         ALGO(file)
    """
    h = hashlib.new(algo)
    with open(filename, 'rb') as f:
        while True:
            data = f.read(h.block_size)
            if not data:
                break
            h.update(data)
    return h.hexdigest()
hashlib.hash_file = hash_file


# this binds new file hashing functions to the hashlib for each existing hash algorithm
for algo in [x for x in hashlib.__dict__.keys()]:
    try:
        h = hashlib.new(algo)
        h.update(b(""))
        def _hash_file(a):
            def _wrapper(f):
                return hash_file(f, a)
            return _wrapper
        setattr(hashlib, "{}_file".format(algo), _hash_file(algo))
    except ValueError:  # unsupported hash type
        pass


class LookupTable(dict):
    """
    Lookup table class for password cracking.

    :param dict_path:   path of the dictionary file to be loaded
    :param algorithm:   the hash algorithm to be used
    :param ratio:       ratio of value to be hashed in the lookup table (by default, every value is considered but, i.e.
                         with a big wordlist, a ratio of 2/3/4/5/... can be used in order to limit the memory load)
    :param dict_filter: function aimed to filter the words from the dictionary (e.g. only alpha-numeric)
    :param prefix:      prefix to be prepended to passwords (e.g. a salt)
    :param suffix:      suffix to be appended to passwords (e.g. a salt)
    """
    def __init__(self, dict_path, algorithm="md5", ratio=1., dict_filter=None, prefix=None, suffix=None):
        dict_path = expanduser(dict_path)
        if not isfile(dict_path):
            raise ValueError("Bad dictionary file path")
        if algorithm not in hashlib.algorithms_available:
            raise ValueError("Bad hashing algorithm")
        if not isinstance(ratio, float) or ratio <= 0. or ratio > 1.:
            raise ValueError("Bad ratio")
        h = lambda x: getattr(hashlib, algorithm)(b(prefix or "") + b(x) + b(suffix or "")).hexdigest()
        with open(dict_path) as f:
            self.filtered = 0
            m = round(1 / float(ratio))
            for i, l in enumerate(f):
                if (i - self.filtered) % m == 0:
                    l = l.strip()
                    if dict_filter is None or hasattr(dict_filter, '__call__') and dict_filter(l):
                        self[h(l)] = l
                    else:
                        self.filtered += 1


hashlib.LookupTable = LookupTable




# -*- coding: UTF-8 -*-
"""Module for enhancing virtualenv preimport.

"""
import os
import re
import site
import sys
import virtualenv
from importlib import import_module
from pip._internal.cli.main_parser import parse_command
from pip._internal.exceptions import PipError
from shutil import rmtree
from six import string_types
from slugify import slugify
from subprocess import Popen, PIPE

from ..helpers import Path, JYTHON, PYPY, WINDOWS


__ORIGINAL_PATH        = os.environ['PATH']
__ORIGINAL_SYSPATH     = sys.path[:]
__ORIGINAL_SYSPREFIX   = sys.prefix
__ORIGINAL_SYSRPREFIX  = getattr(sys, "real_prefix", None)
# package regex for parsing a line with [package]-[version] format
PREGEX = re.compile(r"\s?([a-zA-Z\-_]+?)\-((?:\d+)(?:\.?[a-z]*\d+)+)")


def __activate(venv_dir):
    """
    This activates a virtual environment.
    
    :param venv_dir: virtual environment's directory
    """
    venv_dir = os.path.abspath(venv_dir)
    if not os.path.isdir(venv_dir):
        raise NotAVirtualEnv("Bad virtual environment")
    j = os.path.join
    bin_dir = j(venv_dir, "bin")
    _ = __ORIGINAL_PATH.split(os.pathsep)
    os.environ["PATH"] = os.pathsep.join([bin_dir] + _)
    os.environ["VIRTUAL_ENV"] = venv_dir
    site_packages = j(venv_dir, "Lib", "site-packages") if JYTHON else \
                    j(venv_dir, "site-packages") if PYPY else \
                    j(venv_dir, "Lib", "site-packages") if WINDOWS else \
                    j(venv_dir, "lib", "python{}.{}".format(*sys.version_info), "site-packages")
    old = set(__ORIGINAL_SYSPATH)
    site.addsitedir(site_packages)
    new = list(sys.path)
    sys.path = [i for i in new if i not in old] + [i for i in new if i in old]
    sys.real_prefix = __ORIGINAL_SYSRPREFIX or __ORIGINAL_SYSPREFIX
    sys.prefix = venv_dir


def __check_pip_req_tracker():
    """
    This checks if the temporary folder of the Pip requirements tracker still exists and corrects the related
     environment variable accordingly.
    """
    pip_reqt = os.environ.get('PIP_REQ_TRACKER')
    if pip_reqt is not None and not os.path.exists(pip_reqt):
        os.environ['PIP_REQ_TRACKER'] = ""


def __deactivate():
    """
    This deactivates a virtual environment.
    
    :param venv_dir: virtual environment's directory
    """
    # reset all values modified by activate_this.py
    os.environ['PATH']        = __ORIGINAL_PATH
    os.environ['VIRTUAL_ENV'] = ""
    sys.path                  = __ORIGINAL_SYSPATH[:]
    sys.prefix                = __ORIGINAL_SYSPREFIX
    try:
        # from a normal environment, this key should not exist
        delattr(sys, "real_prefix")
        # but it is important to also reset it if the script was itself run from a virtual environment
        if __ORIGINAL_SYSRPREFIX is not None:
            sys.real_prefix = __ORIGINAL_SYSRPREFIX
    except:
        pass
    __check_pip_req_tracker()


def __get_virtualenv(error=True):
    """
    This gets the currently defined virtual environment or raises an error if no environment is defined.
    """
    venv = os.environ.get('VIRTUAL_ENV', "")
    if venv == "" and error:
        raise NotAVirtualEnv("Not in a virtual environment")
    return venv


def __install(package, *args, **kwargs):
    """
    Python2/3-compatible Pip install function.
    
    :param package: package name
    :param args:    options to be used with the pip install command
    :param kwargs:  keyword-arguments to be used with the pip install command
    """
    global pip_proc
    __check_pip_req_tracker()
    if "-v" in args or "--verbose" in args:
        verbose = True
    else:
        verbose = False
        args += ("-v", )
    error = kwargs.pop("error", False)
    cmd = ["install", "-U"] + __parse_args(*args, **kwargs) + [package.strip()]
    for l in __pip_run(cmd):
        if verbose:
            print(l)
        if not error:
            continue
        if l.startswith("pip._internal.exceptions") or l.startswith("DistributionNotFound"):
            pip_proc.kill()
            raise PipError(l.split(": ", 1)[1])
        elif l.startswith("Successfully installed"):
            m = list(filter(lambda p: p[0] == package, PREGEX.findall(l[23:])))
            if len(m) == 0:
                raise PipError("Installation of {} failed".format(package))
    return PipPackage(package, error)


def __is_installed(package, *args, **kwargs):
    """
    This checks if a given package is installed in the virtual environment.
    
    :param package: package name
    :param args:    options to be used with the pip list command
    :param kwargs:  keyword-arguments to be used with the pip list command
    """
    found = False
    for name, version in __list_packages(*args, **kwargs):
        if isinstance(package, string_types) and package == name or isinstance(package, (list, tuple, set)) and \
           package[0] == name and package[1] == version:
            found = True
            break
    return found


def __list_packages(*args, **kwargs):
    """
    This lists the packages installed in the currently activated or the given virtual environment.
    
    :param venv_dir: virtual environment's directory
    :param args:     options to be used with the pip list command
    :param kwargs:   keyword-arguments to be used with the pip list command
    """
    cmd = ["list"] + __parse_args(*args, **kwargs)
    for line in __pip_run(cmd, False):
        if not ("Package" in line and "Version" in line or "-------" in line or line.strip() == ""):
            yield tuple(_.strip() for _ in line.split(" ", 1))


def __parse_args(*args, **kwargs):
    """
    This parses input args and kwargs in a format that suits pip._internal.main.
    """
    l = []
    for v in args:
        if v not in l:
            l.append(str(v))
    for k, v in kwargs.items():
        k = "--" + k.replace("_", "-")
        if k not in l:
            l.append(k)
            if v is not True:
                l.append(str(v))
    return l


def __pip_run(cmd, error=True):
    """
    This runs a Pip command using the binary from the current virtual environment.
    
    :param cmd: the Pip command and its parameters as a list
    """
    global pip_proc
    venv = __get_virtualenv(error)
    parse_command(cmd)
    cmd = ["pip" if venv == "" else os.path.join(venv, "bin", "pip")] + cmd
    pip_proc = Popen(cmd, stdout=PIPE, stderr=PIPE, universal_newlines=True)
    for line in iter(pip_proc.stdout.readline, ""):
        yield line[:-1]


def __setup(venv_dir, requirements=None, verbose=False):
    """
    This creates (if relevant) and activates a virtual environment. It also allows to define requirements to be
     installed in this environment.
    
    :param venv_dir:     virtual environment's directory
    :param requirements: list of required package OR path of the requirements file to be used
    :param verbose:      displayed Pip output while installing packages
    """
    __deactivate()
    venv_dir = os.path.abspath(venv_dir)
    if not os.path.exists(venv_dir):
        # fix to issue: https://github.com/pypa/virtualenv/issues/1585
        try:
            virtualenv.create_environment(venv_dir)
        except AttributeError:
            virtualenv.cli_run([venv_dir])
    __activate(venv_dir)
    if isinstance(requirements, string_types):
        with open(requirements) as f:
            requirements = [l.strip() for l in f]
    if isinstance(requirements, (tuple, list, set)):
        args = ["-v"] if verbose else []
        kwargs = {'prefix': venv_dir}
        for req in requirements:
            pkg = __install(req, *args, **kwargs)
            for tl in pkg.top_level:
                if hasattr(virtualenv, tl):
                    raise TopLevelAlreadyExists("{} ({})".format(tl, pkg.name))
                m = import_module(tl)
                setattr(virtualenv, tl, m)


def __teardown(venv_dir=None):
    """
    This deactivates and removes the given virtual environment or the one defined in the related environment variable.

    :param venv_dir: virtual environment's directory
    """
    venv = venv_dir or os.environ.get('VIRTUAL_ENV', "")
    if venv != "":
        __deactivate()
        rmtree(venv, True)


class PipPackage(object):
    """
    This class is used to represent a Pip package and its attribute.
    """
    def __init__(self, name, error=True):
        self.name = name
        self.top_level = []
        cmd = ["show", name]
        venv = os.environ.get('VIRTUAL_ENV', "")
        for i in range(2):
            for line in globals()['__pip_run'](cmd, False):
                try:
                    k, v = line.strip().split(": ")
                    k = slugify(k).replace("-", "_")
                    if k.startswith("require"):
                        v = v.split(", ")
                    setattr(self, k, v)
                except ValueError:
                    continue
            if hasattr(self, "version") or venv == "":
                break
            else: # after the first guess, if the package was not found, deactivate it to retry with main Pip
                globals()['__deactivate']()
        # afterwards, re-enable the virtual environment if relevant
        if venv != "":
            globals()['__activate'](venv)
        if not hasattr(self, "version"):
            if error:
                raise PipError("Pip package not found: {}".format(name))
        else:
            name = "{}-{}*".format(self.name, self.version)
            try:
                pkgloc = list(Path(self.location).find(name))[0]
                tl = list(pkgloc.find("top_level.txt"))[0]
                self.top_level = tl.read_lines()
            except IndexError:
                pass


class VirtualEnv(object):
    """
    This context manager simplifies the use of a virtual environment.
    
    :param venvdir:      virtual environment's directory
    :param requirements: list of required package OR path of the requirements file to be used
    :param remove:       whether the virtual environment is to be removed after the entered context
    :param verbose:      displayed Pip output while installing packages
    """
    def __init__(self, venvdir=None, requirements=None, remove=False, verbose=False):
        self.__remove = remove
        self.__requirements = requirements
        self.__venv_dir = venvdir or __get_virtualenv()
        self.__verbose = verbose

    def __enter__(self):
        self.setup(self.__venv_dir, self.__requirements, self.__verbose)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        getattr(self, ["deactivate", "teardown"][self.__remove])()
    
    def __getattr__(self, name):
        try:
            return getattr(virtualenv, name)
        except AttributeError:
            raise AttributeError("object 'VirtualEnv' has no attribute '{}'".format(name))


class NotAVirtualEnv(Exception):
    pass


class TopLevelAlreadyExists(Exception):
    pass


virtualenv.activate      = __activate
virtualenv.deactivate    = __deactivate
virtualenv.install       = __install
virtualenv.is_installed  = __is_installed
virtualenv.list_packages = __list_packages
virtualenv.setup         = __setup
virtualenv.teardown      = __teardown
virtualenv.PipPackage    = PipPackage
virtualenv.VirtualEnv    = VirtualEnv




# -*- coding: UTF-8 -*-
"""Common constants.

"""
import sys
from locale import getlocale
from platform import system


__all__ = __features__ = ["DARWIN", "LINUX", "WINDOWS"]
DARWIN  = system() == "Darwin"
LINUX   = system() == "Linux"
WINDOWS = system() == "Windows"

__all__ += ["ENCODING", "LANGUAGE"]
LANGUAGE, ENCODING = getlocale()

__all__ += ["IPYTHON", "JUPYTER", "JYTHON", "PYPY", "PYTHON3"]
JYTHON  = sys.platform.startswith("java")
PYPY    = hasattr(sys, "pypy_version_info")
PYTHON3 = sys.version_info > (3,)
IPYTHON = JUPYTER = True
try:
    __IPYTHON__
except NameError:
    IPYTHON = JUPYTER = False




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Module for defining interactive mode logic.

"""
import readline
import socket
import sys
from code import compile_command, interact as base_interact, \
                 InteractiveConsole as BaseInteractiveConsole


__all__ = ["set_interact_items"]


def set_interact_items(glob):
    """
    This function prepares the interaction items for inclusion in main script's
     global scope.
    
    :param glob: main script's global scope dictionary reference
    """
    a, l = glob['args'], glob['logger']
    enabled = getattr(a, a._collisions.get("interact") or "interact", False)
    if enabled:
        readline.parse_and_bind('tab: complete')

    # InteractiveConsole as defined in the code module, but handling a banner
    #  using the logging of tinyscript
    class InteractiveConsole(BaseInteractiveConsole, object):
        def __init__(self, banner=None, namespace=None, filename='<console>',
                     exitmsg=None):
            if enabled:
                self.banner = banner
                self.exitmsg = exitmsg
                ns = glob
                ns.update(namespace or {})
                super(InteractiveConsole, self).__init__(locals=ns,
                                                         filename=filename)
            
        def __enter__(self):
            if enabled and self.banner is not None:
                l.interact(self.banner)
            return self
        
        def __exit__(self, *args):
            if enabled and self.exitmsg is not None:
                l.interact(self.exitmsg)
        
        def interact(self, *args, **kwargs):
            if enabled:
                super(InteractiveConsole, self).interact(*args, **kwargs)

    glob['InteractiveConsole'] = InteractiveConsole
    
    def interact(banner=None, readfunc=None, namespace=None, exitmsg=None):
        if enabled:
            if banner is not None:
                l.interact(banner)
            ns = glob
            ns.update(namespace or {})
            base_interact(readfunc=readfunc, local=ns)
            if exitmsg is not None:
                l.interact(exitmsg)

    glob['interact'] = interact
    
    glob['compile_command'] = compile_command if enabled else \
                              lambda *a, **kw: None

    # ConsoleSocket for handling duplicating std*** to a socket for the
    #  RemoteInteractiveConsole
    host = getattr(a, a._collisions.get("host") or "host", None)
    port = getattr(a, a._collisions.get("port") or "port", None)

    # custom socket, for handling the bindings of stdXXX through a socket
    class ConsoleSocket(socket.socket):
        def readline(self, nbytes=2048):
            return self.recv(nbytes)

        def write(self, *args, **kwargs):
            return self.send(*args, **kwargs)

    # RemoteInteractiveConsole as defined in the code module, but handling 
    #  interaction through a socket
    class RemoteInteractiveConsole(InteractiveConsole):
        def __init__(self, *args, **kwargs):
            if enabled:
                # open a socket
                self.socket = ConsoleSocket()
                self.socket.connect((str(host), port))
                # save STDIN, STDOUT and STDERR
                self.__stdin = sys.stdin
                self.__stdout = sys.stdout
                self.__stderr = sys.stderr
                # rebind STDIN, STDOUT and STDERR to the socket
                sys.stdin = sys.stdout = sys.stderr = self.socket
                # now initialize the interactive console
                super(RemoteInteractiveConsole, self).__init__(*args, **kwargs)
        
        def __exit__(self, *args):
            if enabled:
                super(RemoteInteractiveConsole, self).__exit__(*args)
                self.socket.close()
                self.close()
            
        def close(self):
            if enabled:
                # restore STDIN, STDOUT and STDERR
                sys.stdin = self.__stdin
                sys.stdout = self.__stdout
                sys.stderr = self.__stderr

    glob['RemoteInteractiveConsole'] = RemoteInteractiveConsole



# -*- coding: UTF-8 -*-
"""Common checking functions and argument types.

"""
import inspect
import types
from six import integer_types

from .strings import _str2list


__all__ = __features__ = []


# various object type check functions
__all__ += ["is_bool", "is_dict", "is_int", "is_list", "is_neg_int", "is_pos_int"]
is_bool      = lambda b: isinstance(b, bool)
is_dict      = lambda d: isinstance(d, dict)
is_int       = lambda i: isinstance(i, integer_types)
is_list      = lambda l: isinstance(l, (list, set, tuple))
is_neg_int   = lambda i, zero=False: is_int(i) and (i <= 0 if zero else i < 0)
is_pos_int   = lambda i, zero=True: is_int(i) and (i >= 0 if zero else i > 0)

__all__ += ["is_class", "is_coroutine", "is_coroutinefunc", "is_frame", "is_function", "is_generator",
            "is_generatorfunc", "is_instance", "is_lambda", "is_method", "is_module", "is_type"]
is_class         = lambda c: inspect.isclass(c)
is_coroutine     = lambda c: inspect.iscoroutine(c)
is_coroutinefunc = lambda c: inspect.iscoroutinefunction(c)
is_frame         = lambda f: isinstance(f, types.FrameType)
is_function      = lambda f, builtin=False: isinstance(f, getattr(types, ["", "Builtin"][builtin] + "FunctionType"))
is_generator     = lambda g: inspect.isgenerator(g)
is_generatorfunc = lambda g: inspect.isgeneratorfunction(g)
is_instance      = lambda i, cls=None: isinstance(i, cls or object)
is_lambda        = lambda l: isinstance(l, types.LambdaType)
is_method        = lambda m, builtin=False: isinstance(m, getattr(types, ["", "Builtin"][builtin] + "MethodType"))
is_module        = lambda m: isinstance(m, types.ModuleType)
is_type          = lambda t: isinstance(t, type)


# -------------------- DATA FORMAT ARGUMENT TYPES --------------------
__all__ += ["neg_int", "negative_int", "pos_int", "positive_int", "ints", "neg_ints", "negative_ints", "pos_ints",
            "positive_ints"]


def __ints(l, check_func=lambda x: False, idescr=None, **kwargs):
    """ Parses a comma-separated list of ints. """
    l = _str2list(l)
    msg = "{} {}integer{}".format(["Bad list of", "Not a"][len(l) == 1], "" if idescr is None else idescr + " ",
                                  ["s", ""][len(l) == 1])
    if not all(check_func(_, **kwargs) for _ in l):
        raise ValueError(msg)
    return l
ints = lambda l: __ints(l, is_int)
negative_int = neg_int = lambda i, zero=False: __ints(i, is_neg_int, "negative", zero=zero)[0]
positive_int = pos_int = lambda i, zero=True: __ints(i, is_pos_int, "positive", zero=zero)[0]
negative_ints = neg_ints = lambda l, zero=False: __ints(l, is_neg_int, "negative", zero=zero)
positive_ints = pos_ints = lambda l, zero=True: __ints(l, is_pos_int, "positive", zero=zero)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Module for defining stepping mode logic.

"""
import logging
from .helpers.inputs import std_input


__all__ = ["set_step_items"]


COLOR = logging.STEP_COLOR


def set_step_items(glob):
    """
    This function prepares the stepping items for inclusion in main script's
     global scope.
    
    :param glob: main script's global scope dictionary reference
    """
    a = glob['args']
    l = glob['logger']
    enabled = getattr(a, a._collisions.get("step") or "step", False)
    # Step context manager, for defining a block of code that can be paused at
    #  its start and end
    class Step(object):
        def __init__(self, message=None, at_end=False):
            self.message = message
            self.at_end = at_end
            
        def __enter__(self):
            if enabled:
                if self.message:
                    l.step(self.message)
                if not self.at_end:
                    std_input("Press enter to continue", ["bold", COLOR])
                return self
        
        def __exit__(self, *args):
            if enabled and self.at_end:
                std_input("Press enter to continue", ["bold", COLOR])
    glob['Step'] = Step
    # stepping function, for stopping the execution and displaying a message if
    #  any defined
    def step(message=None):
        if enabled:
            if message:
                l.step(message)
            std_input("Press enter to continue", ["bold", COLOR])
    glob['step'] = step



# -*- coding: UTF-8 -*-
"""Module for enhancing logging preimport.

"""
import coloredlogs
import logging
import sys
from functools import wraps

from .inspectp import inspect


PY3 = sys.version[0] == "3"


# setup a null logger
logging.nullLogger = logging.getLogger("null")
logging.nullLogger.setLevel(1000)
logging.nullLogger.addHandler(logging.NullHandler())


def __del(d, k):
    try:
        if isinstance(d, dict):
            del d[k]
        else:
            delattr(d, str(k))
    except (AttributeError, KeyError):
        pass


def addLogLevel(levelName, color, level, bold=True):
    """
    Add a new log level.
    
    :param levelName: name for the new level
    :param color:     related message color
    :param level:     integer defining the level
    :param bold:      whether the related messages should be displayed in bold
    """
    n, N = levelName, levelName.upper()
    if hasattr(logging, N):
        raise ValueError("Cannot overwrite log level '{}'".format(n))
    setattr(logging, N, level)
    setattr(logging, N + "_COLOR", color)
    logging.addLevelName(level, N)
    def display(self, message, *args, **kwargs):
        if self.isEnabledFor(level):
            self._log(level, message, args, **kwargs)
    display.__name__ = n
    setattr(logging.Logger, n, display)
    attrs = {'color': color}
    if bold:
        # compatibility fix due to a change in coloredlogs from version 14.0
        # see: https://github.com/xolox/python-coloredlogs/issues/82
        try:
            attrs['bold'] = coloredlogs.CAN_USE_BOLD_FONT
        except AttributeError:
            # in coloredlogs from v14, CAN_USE_BOLD_FONT is not present anymore and its flag is set to True everywhere
            #  it appears
            attrs['bold'] = True
    coloredlogs.DEFAULT_LEVEL_STYLES[n] = attrs
    if PY3:
        logging._levelToName[level] = N
        logging._nameToLevel[N] = level
    else:
        logging._levelNames[level] = N
logging.addLogLevel = addLogLevel


def bindLogger(f):
    """
    This decorators allows either to bind a logger to self if f is a method or to bind a logger in the local scope of f
     if it is a function. It tries first to get the logger from kwargs, or then tries to get the logger from caller's
     globals or finally sets a null logger. This way, the logger can be used inside the function without caring to get a
     logger itself.
    
    Inspired from: https://stackoverflow.com/questions/17862185/how-to-inject-variable-into-scope-with-a-decorator
    """
    @wraps(f)
    def _wrapper(*args, **kwargs):
        logger = kwargs.pop('logger', None) or globals().get('logger') or inspect.getmainglobals().get('logger') or \
                 logging.nullLogger
        # if f is a method, bind the logger to self
        if inspect.ismethod(f) or f.__code__.co_varnames[0] == "self":
            args[0].logger = logger
            return f(*args, **kwargs)
        # otherwise, pass the logger through globals
        else:
            glob = f.__globals__
            sentinel = object()
            old = glob.get('logger', sentinel)
            glob['logger'] = logger
            try:
                return f(*args, **kwargs)
            finally:
                if old is sentinel:
                    del glob['logger']
                else:
                    glob['logger'] = old
    return _wrapper
logging.bindLogger = bindLogger


def delLevelName(level):
    """
    Remove association of 'levelName' with 'level'.
    """
    logging._acquireLock()
    if isinstance(level, int):
        levelName = logging._levelToName[level] if PY3 else logging._levelNames[level]
    else:
        levelName = level.upper()
        level = logging._nameToLevel.get(levelName) if PY3 else \
                {v: k for k, v in logging._levelNames.items()}.get(levelName)
    __del(getattr(logging, "_levelToName", None), level)
    __del(getattr(logging, "_levelNames", None), level)
    __del(getattr(logging, "_nameToLevel", None), levelName)
    logging._releaseLock()
logging.delLevelName = delLevelName


def delLogLevel(levelName):
    """
    Remove a log level.
    """
    n, N = levelName, levelName.upper()
    if not hasattr(logging, N):
        raise ValueError("Log level '{}' does not exist".format(n))
    __del(logging, N)
    __del(globals(), "{}_COLOR".format(N))
    delLevelName(levelName)
    __del(logging.Logger, n)
    __del(coloredlogs.DEFAULT_LEVEL_STYLES, n)
logging.delLogLevel = delLogLevel


def lastLogRecord():
    """
    Display the last log record.
    """
    rec = getattr(logging, "_last_record", None)
    if rec:
        lastrec = logging.getLogger("__last_record__")
        lastrec.name = rec.name
        lastrec.log(rec.levelno, rec.msg)
        lastrec.name = "__last_record__"
logging.lastLogRecord = lastLogRecord


def setLogger(name=None):
    """
    Set up the logger with the given name according to Tinyscript's logging configuration.
    
    :param name: logger name
    """
    setLoggers(name)        
logging.setLogger = setLogger


def setLoggers(*names):
    """
    Set up the loggers with the given names according to Tinyscript's logging configuration.
    
    :param names: logger names
    """
    if len(names) == 0:
        names = [None]
    main = inspect.getmainglobals().get('logger') or logging.getLogger()
    main.setLevel(1)
    if not any(type(h) is InterceptionHandler for h in main.handlers):
        main.addHandler(InterceptionHandler())
    # ensure that the main logger has no parent
    main.parent = None
    for name in names:
        logger = logging.getLogger(name)
        # check that the given logger is not Tinyscript's one
        if main.name == name or id(main) == id(logger):
            continue
        # set Tinyscript's logger as the logger's parent
        logger.parent = main
        # copy the reference to the list of handlers
        logger.handlers = main.handlers
        # disable propagation from the sublogger so that it does not duplicate log messages
        logger.propagate = False
logging.setLoggers = setLoggers


class InterceptionHandler(logging.Handler):
    """ Dummy handler saving the last handled log record. """
    def __init__(self):
        super(InterceptionHandler, self).__init__(1)
    
    def handle(self, record):
        logging._last_record = record
logging.InterceptionHandler = InterceptionHandler


# setup the private logger for displaying the last intercepted log record
__logger = logging.getLogger("__last_record__")
__logger.setLevel(1)
__handler = logging.StreamHandler()
__formatter = logging.Formatter('\r%(asctime)s [%(levelname)s] %(message)s')
__handler.setFormatter(__formatter)
__logger.addHandler(__handler)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Module for extending argparse features, for simplifying parser.py.

"""
import random
import re
import shlex
import sys
from argparse import _ActionsContainer, _ArgumentGroup, \
                     _MutuallyExclusiveGroup, _AttributeHolder, \
                     _SubParsersAction, ArgumentDefaultsHelpFormatter, \
                     Action, ArgumentError, RawTextHelpFormatter, SUPPRESS, \
                     _UNRECOGNIZED_ARGS_ATTR, Namespace as BaseNamespace, \
                     ArgumentParser as BaseArgumentParser
from os import environ
from os.path import abspath, basename, dirname, splitext
from stat import S_IXUSR
try:
    from configparser import ConfigParser, NoOptionError, NoSectionError
except ImportError:
    from ConfigParser import ConfigParser, NoOptionError, NoSectionError

from .helpers.constants import PYTHON3
from .helpers.inputs import user_input
from .helpers.licenses import *
from .helpers.data.types import is_executable, is_long_opt, is_pos_int, \
                                is_short_opt
from .helpers.text import *
from .helpers.text import configure_docformat, txt_terminal_render
from .loglib import logger


__all__ = ["ArgumentParser", "DUNDERS", "SCRIPTNAME_FORMAT", "SUPPRESS"]


BASE_DUNDERS = ['__author__', '__copyright__', '__credits__', '__license__',
                '__reference__', '__source__', '__training__']
DUNDERS = BASE_DUNDERS + [
    '__date__', '__details__', '__description__', '__doc__', '__docformat__',
    '__email__', '__examples__', '__functions__', '__maximum_python_version__',
    '__minimum_python_version__', '__priority__', '__product__', '__script__',
    '__status__', '__version__',
]

DEFAULT_MAX_LEN     = 20
DEFAULT_LST_MAX_LEN = 10
SCRIPTNAME_FORMAT   = "slugified"
SCRIPTNAME_FORMATS  = {
    'acronym':   lambda s: "".join(x.strip()[0].upper() for x in \
                           re.split(r"[ -_]", s)) \
                           if len(re.split(r"[ -_]", s)) > 1 else s.upper(),
    'as_is':     lambda s: s,
    'none':      lambda s: s,
    'slugified': lambda s: "".join(x.strip().capitalize() \
                                   for x in re.split(r"[ -_]", s)),
}


# ------------------------------- CUSTOM ACTIONS -------------------------------
class _ConfigAction(Action):
    """
    Custom action for handling an INI configuration file.
    """
    def __init__(self, option_strings, dest=None, default=None, help=None):
        super(_ConfigAction, self).__init__(option_strings=option_strings,
                                            dest=SUPPRESS, default=default,
                                            nargs=1, help=gt(help),
                                            metavar="INI")

    def __call__(self, parser, namespace, values, option_string=None):
        conf = values[0]
        setattr(namespace, "read_config", conf)
        if conf not in parser._config.read(conf):
            logger.error(gt("Config file '{}' not found").format(conf))
            sys.exit(2)
        parser.config_args()


class _DemoAction(Action):
    """
    Custom action for triggering the execution of an example.
    """
    def __init__(self, option_strings, dest=SUPPRESS, help=None):
        super(_DemoAction, self).__init__(option_strings=option_strings,
                                          dest=SUPPRESS, default=SUPPRESS,
                                          nargs=0, help=gt(help))

    def __call__(self, parser, namespace, values, option_string=None):
        parser.demo_args()


class _ExtendAction(Action):
    """
    Custom action for extending a list of values.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        _ = getattr(namespace, self.dest) or []
        if not isinstance(values, list):
            values = [values]
        _.extend(values)
        setattr(namespace, self.dest, _)


class _NewSubParsersAction(_SubParsersAction):
    """
    Modified version of argparse._SubParsersAction for handling formatters of
     subparsers, inheriting from this of the main parser.
    """
    last = False
    
    def __init__(self, *args, **kwargs):
        kwargs.pop('required', True)
        super(_NewSubParsersAction, self).__init__(*args, **kwargs)
    
    def add_parser(self, name, **kwargs):
        # set prog from the existing prefix
        if kwargs.get('prog') is None:
            kwargs['prog'] = "%s %s" % (self._prog_prefix, name)
        # create a pseudo-action to hold the choice help
        aliases = kwargs.pop('aliases', [])
        if 'help' in kwargs:
            help = kwargs.pop('help')
            # see [Python2] argparse.py:1029 and [Python3] argparse.py:1059
            args = (name, aliases, help) if PYTHON3 else (name, help)
            choice_action = self._ChoicesPseudoAction(*args)
            self._choices_actions.append(choice_action)
        # create the parser, but with another formatter and separating the help
        #  into an argument group
        parser = self._parser_class(ArgumentParser.globals_dict, add_help=False,
                                    **kwargs)
        parser.name = name
        i = parser.add_argument_group(txt2title(gt("extra arguments")))
        i.add_argument("-h", action="usage", prefix="show",
                       help=gt("show usage message and exit"))
        i.add_argument("--help", action="help", prefix="show",
                       help=gt("show this help message and exit"))
        # add it to the map
        self._name_parser_map[name] = parser
        # make parser available under aliases also (Python3 only ; see before)
        for alias in aliases:
            self._name_parser_map[alias] = parser
        return parser


class _UsageAction(Action):
    """
    Custom action for displaying the usage message.
    """
    def __init__(self, option_strings, dest=SUPPRESS, help=None):
        super(_UsageAction, self).__init__(option_strings=option_strings,
                                           dest=SUPPRESS, default=SUPPRESS,
                                           nargs=0, help=gt(help))

    def __call__(self, parser, namespace, values, option_string=None):
        parser.print_usage()
        parser.exit()


class _WizardAction(Action):
    """
    Custom action for triggering the wizard, asking for argument values.
    """
    def __init__(self, option_strings, dest=SUPPRESS, help=None):
        super(_WizardAction, self).__init__(option_strings=option_strings,
                                            dest=SUPPRESS, default=SUPPRESS,
                                            nargs=0, help=gt(help))

    def __call__(self, parser, namespace, values, option_string=None):
        parser.input_args()


# ------------------------------ CUSTOM ENTITIES -------------------------------
class _NewActionsContainer(_ActionsContainer):
    """
    Modified version of argparse._ActionsContainer for handling a new "note"
     keyword argument.
    """
    def __init__(self, *args, **kwargs):
        super(_NewActionsContainer, self).__init__(*args, **kwargs)
        self.register('action', 'parsers', _NewSubParsersAction)
        self.register('action', 'config', _ConfigAction)
        self.register('action', 'demo', _DemoAction)
        self.register('action', 'extend', _ExtendAction)
        self.register('action', 'usage', _UsageAction)
        self.register('action', 'wizard', _WizardAction)
    
    def add_argument(self, *args, **kwargs):
        new_kw = {k: v for k, v in kwargs.items()}
        # collect Tinyscript-added keyword-arguments
        cancel = new_kw.pop('cancel', False)
        orig = new_kw.pop('orig', None)
        note = new_kw.pop('note', None)
        last = new_kw.pop('last', False)
        prefix = new_kw.pop('prefix', None)
        suffix = new_kw.pop('suffix', None)
        try:
            # define the action based on argparse, with only argparse-known
            #  keyword-arguments
            action = super(_NewActionsContainer, self).add_argument(*args,
                                                                    **new_kw)
            # now set Tinyscript-added keyword-arguments
            action.note = None if note is None else gt(note)
            action.last = last
            action.orig = orig
            action.prefix = prefix
            action.suffix = suffix
            return args[-1]
        except ArgumentError:
            # drop the argument if conflict and cancel set to True
            if cancel:
                return
            # otherwise, retry after removing the short option string
            args = list(args)
            short_opt = list(filter(is_short_opt, args))
            if len(short_opt) > 0:
                args.remove(short_opt[0])
                if len(args) > 0:
                    return self.add_argument(*args, **kwargs)
            # otherwise, retry after modifying the long option string with the
            #  precedence to the prefix (if set) then the suffix (if set)
            long_opt = list(filter(is_long_opt, args))
            if len(long_opt) > 0:
                long_opt = args.pop(args.index(long_opt[0]))
                if kwargs.get('action') in \
                    [None, 'store', 'append', 'store_const', 'append_const']:
                    # set metavar only if no choices given ; otherwise, it
                    #  takes the precedence on choices in the help message
                    kwargs['metavar'] = kwargs.get('metavar') or \
                                        (long_opt.lstrip('-').upper() \
                                         if not kwargs.get('choices') else None)
                curr_opt = long_opt.lstrip("-")
                kwargs['orig'] = curr_opt.replace("-", "_")
                if prefix:
                    long_opt = "--{}-{}".format(prefix, curr_opt)
                    args.append(long_opt)
                    return self.add_argument(*args, **kwargs)
                elif suffix:
                    long_opt = "{}-{}".format(long_opt, suffix)
                    args.append(long_opt)
                    return self.add_argument(*args, **kwargs)

    def add_argument_group(self, *args, **kwargs):
        group = _NewArgumentGroup(self, *args, **kwargs)
        self._action_groups.append(group)
        return group

    def add_mutually_exclusive_group(self, **kwargs):
        group = _NewMutuallyExclusiveGroup(self, **kwargs)
        self._mutually_exclusive_groups.append(group)
        return group


class _NewArgumentGroup(_ArgumentGroup, _NewActionsContainer):
    """
    Modified version of argparse._ArgumentGroup for modifying arguments groups
     handling in the modified ActionsContainer.
    """
    pass
        

class _NewMutuallyExclusiveGroup(_MutuallyExclusiveGroup, _NewArgumentGroup):
    """
    Modified version of argparse._MutuallyExclusiveGroup for modifying arguments
     mutually exclusive groups handling in the modified ActionsContainer.
    """
    pass
        

class ArgumentParser(_NewActionsContainer, BaseArgumentParser):
    """
    Modified version of argparse.ArgumentParser, based on the modified
     ActionsContainer.
    
    :param globals_dict: globals() dictionary from the calling script/tool
                         NB: this is only for help formatting purpose ;
                             therefore this is NOT propagated through subparsers
    :param args:         arguments applicable for argparse.ArgumentParser
    :param kwargs:       kwarguments applicable for argparse.ArgumentParser
                         NB: 'prog' set to preformatted program name
                             'add_help' set to False (customized)
                             'conflict_handler' set to "error" (cfr tuning of
                                                                add_argument)
                             'formatter_class' set to custom HelpFormatter
                             'epilog' set to preformatted usage message
                             'description' set to preformatted help message
    """
    _config = ConfigParser()
    is_action = lambda s, a, *l: any(type(a) is s._registry_get('action', n) \
                                     for n in l)
    name = "main"
    
    def __init__(self, globals_dict=None, *args, **kwargs):
        ArgumentParser.globals_dict = gd = globals_dict or {}
        configure_docformat(gd)
        self._config_parsed = False
        self._docfmt = gd.get('__docformat__')
        self._reparse_args = {'pos': [], 'opt': [], 'sub': []}
        self.examples = gd.get('__examples__', [])
        script = gd.get('__file__', sys.argv[0])
        if script and kwargs.get('prog') is None:
            path = abspath(script)
            root = dirname(path)
            script = basename(script)
            kwargs['prog'] = "python{} ".format(["", "3"][PYTHON3]) + \
                          script if not is_executable(path) else "./" + script \
                          if root not in environ['PATH'].split(":") else script
            ArgumentParser.prog = kwargs['prog']
            script, _ = splitext(script)
        kwargs['add_help'] = False
        kwargs['conflict_handler'] = "error"
        kwargs['formatter_class'] = HelpFormatter
        # format the epilog message
        if self.examples and script:
            _ = ["{} {}".format(ArgumentParser.prog, e) for e in self.examples]
            _ = list(filter(lambda x: x.startswith(kwargs['prog']), _))
            if len(_) > 0:
                kwargs['epilog'] = txt2title(gt("Usage example{}"
                                   .format(["", "s"][len(_) > 1])) + ":")
                e = '\n'.join(["\n", "  "][self._docfmt is None] + \
                              txt2paragraph(e) for e in _)
                kwargs['epilog'] += "\n" + e
        # adapt the script name ; if SCRIPTNAME is provided, it supersedes
        #  SCRIPTNAME_FORMAT, otherwise compute the name according to the
        #  format specified in SCRIPTNAME_FORMAT
        sname = gd.get('__script__')
        if sname is None:
            sname_fmt  = gd.get('SCRIPTNAME_FORMAT', SCRIPTNAME_FORMAT)
            sname_func = SCRIPTNAME_FORMATS.get(sname_fmt)
            if sname_func:
                sname = sname_func(script)
            else:
                l = "\n- ".join(sorted(SCRIPTNAME_FORMATS.keys()))
                raise ValueError("Bad script name format ; please use one of "
                                 "the followings:\n{}".format(l))
        self.scriptname = gd['__scriptname__'] = sname
        # format the description message
        d = sname
        d += " " + str(gd.get('__version__') or "")
        d = txt2title(d, level=1)
        v = gd.get('__status__')
        if v:
            d += " (" + v + ")"
        l = max(list(map(lambda x: len(txt2italic(x.strip('_'))),
                         BASE_DUNDERS)))
        for k in BASE_DUNDERS:
            m = gd.get(k)
            if m:
                if k == '__copyright__':
                    if not isinstance(m, tuple):
                        m = (m, )
                    m = copyright(*m)
                elif k == '__license__':
                    m = license(m, True) or m
                meta = ("{: <%d}: {}" % l) \
                       .format(txt2italic(k.strip('_').capitalize()), m)
                if k == '__author__':
                    e = gd.get('__email__')
                    if e:
                        meta += " ({})".format(txt2email(e))
                d += ["\n\n", "\n"][self._docfmt is None] + txt2paragraph(meta)
        doc = txt2blockquote(gd.get('__doc__') or "")
        if doc:
            d += "\n\n" + doc
        kwargs['description'] = d
        self.details = gd.get('__details__', [])
        # now initialize argparse's ArgumentParser with the new arguments
        super(ArgumentParser, self).__init__(*args, **kwargs)
    
    def _filtered_actions(self, *a_types):
        """
        Get actions filtered on a list of action types.
        
        :param a_type: argparse.Action instance name (e.g. count, append)
        """
        for a in filter(lambda _: self.is_action(_, *a_types), self._actions):
            yield a
    
    def _input_arg(self, a):
        """
        Ask the user for input of a single argument.
        
        :param a: argparse.Action instance
        :return:  the user input, asked according to the action
        """
        # if action of an argument that suppresses any other, just return
        if a.dest == SUPPRESS or a.default == SUPPRESS:
            return
        # prepare the prompt
        prompt = gt((a.help or a.dest).capitalize())
        r = {'newline': True, 'required': a.required}
        # now handle each different action
        if self.is_action(a, 'store', 'append'):
            return user_input(prompt, a.choices, a.default, **r)
        elif self.is_action(a, 'store_const', 'append_const'):
            return user_input(prompt, (gt("(A)dd"), gt("(D)iscard")), "d", **r)
        elif self.is_action(a, 'store_true'):
            return user_input(prompt, (gt("(Y)es"), gt("(N)o")), "n", **r)
        elif self.is_action(a, 'store_false'):
            return user_input(prompt, (gt("(Y)es"), gt("(N)o")), "y", **r)
        elif self.is_action(a, 'count'):
            return user_input(prompt, is_pos_int, 0, gt("positive integer"),
                   **r)
        elif self.is_action(a, 'parsers'):
            pmap = a._name_parser_map
            _ = list(pmap.keys())
            return user_input(prompt, _, _[0], **r) if len(_) > 0 else None
        raise NotImplementedError(gt("Unknown argparse action"))
    
    def _reset_args(self):
        args = [_ for _ in self._reparse_args['pos']] + \
               [_ for _ in self._reparse_args['opt']]
        for sp in self._reparse_args['sub']:
            args += sp._reset_args()
        self._reparse_args = {'pos': [], 'opt': [], 'sub': []}
        return args
    
    def _set_arg(self, a, s="main", c=False):
        """
        Set a single argument.
        
        :param a: argparse.Action instance
        :param s: config section title
        :param c: use class' ConfigParser instance to get parameters
        """
        # if action of an argument that suppresses any other, just return
        if a.dest is SUPPRESS or a.default is SUPPRESS:
            return
        # check if an option string is used for this action in sys.argv ;
        #  if so, simply return as it will be parsed normally
        if any(o in sys.argv[1:] for o in a.option_strings):
            return
        # in case of non-null config, get the value from the config object
        default = a.default if a.default is None else str(a.default)
        if c:
            try:
                value = ArgumentParser._config.get(s, a.dest)
            except (NoOptionError, NoSectionError) as e:
                item = "setting" if isinstance(e, NoOptionError) else "section"
                # if the argument is required, just ask for the value
                value = self._input_arg(a) if a.required else default
                logger.debug(gt("{} {} not present in config (set to {})")
                             .format(a.dest, item, value))
        # in case of null config, just ask for the value
        else:
            value = self._input_arg(a)
        # collect the option string before continuing
        try:
            ostr = a.option_strings[0]
        except IndexError:  # occurs when positional argument
            ostr = None
        # now handle arguments regarding the action
        if self.is_action(a, 'store', 'append'):
            if value:
                if ostr:
                    self._reparse_args['opt'].extend([ostr, value])
                else:
                    self._reparse_args['pos'].extend([value])
        elif self.is_action(a, 'store_const', 'append_const'):
            if value.lower() == "add" or value != default:
                self._reparse_args['opt'].append(ostr)
        elif self.is_action(a, 'store_true'):
            if value.lower() in ["y", "true"]:
                self._reparse_args['opt'].append(ostr)
        elif self.is_action(a, 'store_false'):
            if value.lower() in ["n", "false"]:
                self._reparse_args['opt'].append(ostr)
        elif self.is_action(a, 'count'):
            v = int(value or 0)
            if v > 0:
                if ostr.startswith("--"):
                    new_arg = [ostr for i in range(v)]
                else:
                    new_arg = ["-{}".format(v * ostr.strip('-'))]
                self._reparse_args['opt'].extend(new_arg)
        elif self.is_action(a, 'parsers'):
            if not value:
                value = self._input_arg(a)
            pmap = a._name_parser_map
            if c:
                pmap[value].config_args(a.dest)
                pmap[value]._reparse_args['pos'].insert(0, value)
            else:
                pmap[value].input_args()
            self._reparse_args['sub'].append(pmap[value])
        else:
            raise NotImplementedError("Unknown argparse action")
    
    def _sorted_actions(self):
        """
        Generate the sorted list of actions based on the "last" attribute.
        """
        for a in filter(lambda _: not _.last and \
                        not self.is_action(_, 'parsers'), self._actions):
            yield a
        for a in filter(lambda _: _.last and \
                        not self.is_action(_, 'parsers'), self._actions):
            yield a
        for a in filter(lambda _: self.is_action(_, 'parsers'),
                        self._actions):
            yield a
    
    def config_args(self, section="main"):
        """
        Additional method for feeding input arguments from a config file.
        
        :param section: current config section name
        """
        if self._config_parsed:
            return
        for a in self._filtered_actions("config"):
            for o in a.option_strings:
                try:
                    i = sys.argv.index(o)
                    sys.argv.pop(i)  # remove the option string
                    sys.argv.pop(i)  # remove the value that follows
                except ValueError:
                    pass
        for a in self._sorted_actions():
            self._set_arg(a, section, True)
        self._config_parsed = True
    
    def demo_args(self):
        """
        Additional method for replacing input arguments by demo ones.
        """
        argv = random.choice(self.examples).replace("--demo", "")
        self._reparse_args['pos'] = shlex.split(argv)
    
    def error(self, message):
        """
        Prints a usage message incorporating the message to stderr and exits in
         the case when no new arguments to be reparsed, that is when no special
         action like _DemoAction (triggering parser.demo_args()) or 
         _WizardAction (triggering input_args()) was called. Otherwise, it
         simply does not stop execution so that new arguments can be reparsed.
        """
        if all(len(x) == 0 for x in self._reparse_args.values()):
            # normal behavior with argparse
            self.print_usage(sys.stderr)
            self.exit(2, gt("%s: error: %s\n") % (self.prog, message))
    
    def format_help(self):
        text = ""
        # description
        formatter = self._get_formatter()
        formatter.add_text(self.description)
        text += formatter.format_help() + "\n"
        # usage
        text += self.format_usage().rstrip("\n") + "\n\n"
        # positionals, optionals and user-defined groups
        for action_group in self._action_groups:
            formatter = self._get_formatter()
            formatter.start_section(txt2title(gt(action_group.title) + ":"))
            formatter.add_text(action_group.description)
            formatter.add_arguments(action_group._group_actions)
            formatter.end_section()
            actions = formatter.format_help()
            a = ""
            for i, line in enumerate(actions.splitlines()):
                s, dedent = line.lstrip(), 0
                if i == 0:
                    # action_group.title has ":" added after txt2title(...)
                    #  causing an extra ":" to be appended
                    if self._docfmt == "html":
                        # for markup languages with tags, ":" appears behind
                        #  the title tag
                        a += line.rstrip(":") + "\n"
                    else:
                        # for other markup languages without tags, ":"
                        #  appears at the end in duplicate
                        a += line.rstrip(":") + ":\n"
                    continue
                if i == 1 and self._docfmt == "rst":
                    # action_group.title has ":" added after txt2title(...)
                    #  causing an extra ":" to be appended behind the
                    #  underline, making rendering fail
                    a = a.rstrip("\n") + "\n" + line.rstrip(":") + "\n"
                    continue
                if self._docfmt and s[0] in "-*":
                    dedent = len(line) - len(s)
                a += (line + "\n\n") if line.rstrip()[-1] == ":" else \
                     (txt2paragraph(line[dedent:]) + \
                     ["\n\n", "\n"][self._docfmt is None])
            text += a + "\n"
        # epilog
        formatter = self._get_formatter()
        formatter.add_text(self.epilog)
        text += formatter.format_help()
        # determine help from format above
        return txt_terminal_render(text)
    
    def format_usage(self):
        formatter = self._get_formatter()
        formatter.add_usage(self.usage, self._actions,
                            self._mutually_exclusive_groups)
        text = formatter.format_help()
        if self._docfmt:
            title, usage = text.split(": ", 1)
            text = txt2title(title + ":") + "\n\n" + \
                   txt2paragraph(usage.rstrip("\n")) + "\n"
        return text
    
    def input_args(self):
        """
        Additional method for making the user input arguments manually.
        """
        for a in self._sorted_actions():
            self._set_arg(a)
    
    def parse_args(self, args=None, namespace=None):
        """
        Reparses new arguments when _DemoAction (triggering parser.demo_args())
         or _WizardAction (triggering input_args()) was called.
        """
        if not namespace:  # use the new Namespace class for handling _config
            namespace = Namespace(self)
        if len(sys.argv) == 2 and sys.argv[1] == "DISPLAY_USAGE":
            self.print_usage()
            self.exit()
        namespace = super(ArgumentParser, self).parse_args(args, namespace)
        if len(self._reparse_args['pos']) > 0 or \
           len(self._reparse_args['opt']) > 0 or \
           len(self._reparse_args['sub']) > 0:
            args = self._reset_args()
            namespace = super(ArgumentParser, self).parse_args(args, namespace)
        # process "-hh..." here, after having parsed the arguments
        help_level = getattr(namespace, "help", 0)
        if help_level > 0:
            self.print_help()
            self.print_extended_help(help_level)
            self.exit()
        return namespace
    
    def print_extended_help(self, level=1, file=None):
        if not isinstance(self.details, (tuple, list, set)):
            self.details = [self.details]
        for _, message in zip((level - 1) * [None], self.details):
            message = "\n{}\n".format(message.strip())
            self._print_message(message, file or sys.stdout)
    
    def print_usage(self, file=None):
        self._print_message(txt_terminal_render(self.format_usage()),
                            file or sys.stdout)
    
    @classmethod
    def add_to_config(cls, section, name, value):
        """
        Add a parameter to the shared ConfigParser object.
        
        :param section: parameter's section
        :param name:    parameter's name
        :param value:   parameter's value
        """
        if value:
            if not cls._config.has_section(section):
                cls._config.add_section(section)
            cls._config.set(section, name, str(value))
    
    @classmethod
    def reset(cls):
        cls._config = ConfigParser()


class HelpFormatter(ArgumentDefaultsHelpFormatter, RawTextHelpFormatter):
    """
    Help message formatter for appending a custom note (as input through the
     add_argument method of CustomArgumentParser) to argument help. It also
     allows to reduce long default values (e.g. a list of integers) to something
     readable.
    """
    def _expand_help(self, action):
        params = dict(vars(action), prog=self._prog)
        for name in list(params):
            if params[name] is SUPPRESS:
                del params[name]
        for name in list(params):
            if hasattr(params[name], '__name__'):
                params[name] = params[name].__name__
        if params.get('choices') is not None:
            choices_str = ', '.join([str(c) for c in params['choices']])
            params['choices'] = choices_str
        if params.get('default') is not None:
            s = str(params['default'])
            # if the default value string representation is too long, reduce it
            if len(s) > DEFAULT_MAX_LEN:
                p = s.split(',')
                if len(p) > DEFAULT_LST_MAX_LEN:
                    s = ','.join(p[:2] + ["..."] + p[-2:])
            params['default'] = s
        return self._get_help_string(action) % params

    def _get_help_string(self, action):
        help = super(HelpFormatter, self)._get_help_string(action)
        if '%(note)' not in help and hasattr(action, "note") and \
            action.note is not None:
            help += '\n NB: %(note)s'
        return help


class Namespace(BaseNamespace):
    """
    Modified Namespace class for handling ArgumentParser._config.
    """
    # private __dict__, so that vars() can still be used with no "junk" variable
    #  used for Tinyscript-related processing (e.g. _current_parser)
    __privdict__ = {}
    # exclude list for saving options in a ConfigParser object
    excludes = ["_current_parser", "_debug_level", "_collisions",
                "_subparsers", "read_config", "write_config"]
    
    def __init__(self, parser):
        self._current_parser = parser.name
        self._collisions = {a.orig: a.dest for a in parser._actions \
                            if getattr(a, "orig", None)}
        self._subparsers = [a.dest for a in parser._filtered_actions("parsers")]
    
    def __getattr__(self, name):
        # handle __privdict__ entry first
        if (name.startswith("_") or name in self.excludes) and \
            name in self.__privdict__:
            return self.__privdict__[name]
        # then get the attribute the normal way
        return self.__getattribute__(name)
    
    def __setattr__(self, name, value):
        # handle __privdict__ entry first
        if (name.startswith("_") or name in self.excludes) and \
            name != "_debug_level":
            self.__privdict__[name] = value
        else:
            super(Namespace, self).__setattr__(name, value)
        # then save the entry to the ConfigParser object if not excluded
        if name not in self.excludes:
            ArgumentParser.add_to_config(self._current_parser, name, value)
        # finally switch the current parser value if the option name is part of
        #  a subparser's list of options ; this new name will allow to save new
        #  options in a new section of the ConfigParser object
        if hasattr(self, "_subparsers") and name in self._subparsers and value:
            self._current_parser = name
    
    def get(self, name):
        try:
            return self.__getattr__(name)
        except AttributeError:
            return



# -*- coding: UTF-8 -*-
"""Module for enhancing random preimport.

"""
import operator
import random
import struct

from ..helpers.compat import b
from ..helpers.data import *


class __Base(object):
    """ Class holding some common methods for other classes of this module """
    def next_block(self, output_format="str", update=True):
        """
        Get the next 32bits-block of data.
        
        :param output_format: output format (string, hexadecimal or binary)
        :param update:        update the state of the generator
        """
        return self.next_blocks(1, output_format, update)

    def next_blocks(self, n, output_format="str", update=True):
        """
        Get the n next 32bits-blocks of data.
        
        :param n:             number of blocks of n bits to be generated after the target (from start if target is None)
        :param output_format: output format (string, hexadecimal or binary)
        :param update:        update the state of the generator
        """
        l = len(getattr(self, "target", None) or "")
        s = self.get(l + n * self.n, output_format, update)
        if output_format == "str":
            return s[l//8:]
        elif output_format == "hex":
            return s[l//4:]
        return s[l:]


class Geffe(__Base):
    """
    Geffe generator class.
    
    This can be defined by its key (12-chars string or 96 bits) or by a tuple of 3 seeds for its internal LFSR.

    :param key:    12-chars key or str/list of 96 bits
    :param seeds:  3-int tuple (or list) with the seeds for the LFSR
    :param update: update the state of the generator when bits are generated

    Example usage:

      >>> from tinyscript import *
      >>> g = random.Geffe("a_12bits_key")
      >>> print(g.next_block())
    """
    n = 32
    taps = [
        (3, 4, 6, 7, 8, 9, 11, 32),
        (1, 2, 3, 4, 6, 8, 9, 13, 14, 32),
        (1, 2, 3, 7, 12, 14, 15, 32),
    ]
    
    def __init__(self, key=None, seeds=(0, 0, 0)):
        seeds = Geffe._format_param(key, seeds)
        self._lfsrs = [LFSR(s, t, 32) for s, t in zip(seeds, Geffe.taps)]
        self.target = self._lfsrs[0].target
    
    def get(self, length=None, output_format="str", update=False):
        """
        Get a given number of bits from the Geffe generator with the given output format.
        
        :param length:        number of bits to be generated
        :param output_format: output format (string, hexadecimal or binary)
        :param update:        update the state of the generator
        """
        if output_format not in ["str", "hex", "bin"]:
            raise ValueError("Bad output format")
        stream = []
        lfsr_get = lambda i: self._lfsrs[i].get(length, "bin", update)
        for bits in zip(lfsr_get(0), lfsr_get(1), lfsr_get(2)):
            stream.append(Geffe.F(*bits))
        if output_format == "str":
            return bin2str(stream)
        elif output_format == "hex":
            return bin2hex(stream)
        else:
            return stream
        
    @staticmethod
    def F(x1, x2, x3):
        return (x1 & x2) ^ (int(not x1) & x3)
    
    @staticmethod
    def _format_param(key, seeds):
        if key is not None:
            l = len(key)
            if is_str(key):
                if l <= 12:
                    seeds = struct.unpack("iii", b(pad(key, "incremental", 12)[:12]))
                elif l == 96 and is_bin(key):
                    seeds = tuple([bin2int(key[i:i+32]) for i in range(0, 96, 32)])
                else:
                    raise ValueError("Bad key length ({}) ; should be 12".format(l))
            elif is_list(key):
                if l == 3 and all(is_str(s) for s in key):
                    seeds = tuple(map(lambda s: struct.unpack("i", b(s).ljust(4, b"\x00"))[0], key))
                elif l == 96 and is_bin(key):
                    bits = ''.join(list(map(str, key)))
                    seeds = tuple(map(lambda x: int("0b" + x, 2), [bits[i:i+32] for i in range(0, 96, 32)]))
        if len(seeds) != 3 or any(not is_pos_int(i, False) or i.bit_length() > 32 for i in seeds):
            raise ValueError("Invalid seeds (should be a 3-tuple of 32-bits integers)")
        return seeds


class LFSR(__Base):
    """
    Linear Feedback Shift Register (LFSR) class.
    
    This can be defined by its parameters (seed, taps, nbits) or can be determined from a target using the
     Berlekamp-Massey algorithm.

    :param seed:   integer, list of bits or string for initializing the LFSR
    :param taps:   tuple or list of indices of the polynomial
    :param nbits:  LFSR register length of bits
    :param target: input string or list of bits to be matched while determining LFSR's parameters
    :param update: update the state of the LFSR when bits are generated

    Example usage:

      >>> from tinyscript import *
      >>> l = random.LFSR("0123456789abcdef")
      >>> print(l.next_block())
    """
    def __init__(self, seed=0, taps=None, nbits=None, target=None):
        if target is not None and (taps is None or nbits is None):
            self.target = LFSR._format_target(target)
            self.__berlekamp_massey_algorithm()
        else:
            self.seed, self.taps, self.n = LFSR._format_param(seed, taps, nbits)
            self.target = [b for b in self.seed]
        # test if parameters are correct
        self.test()
    
    def __berlekamp_massey_algorithm(self):
        """
        Berlekamp-Massey algorithm for finding the shortest LFSR for a given binary output sequence.
         
        See: https://en.wikipedia.org/wiki/Berlekamp%E2%80%93Massey_algorithm
        """
        bs = list(map(int, [b for b in self.target]))
        n = len(bs)
        b, c = [0 for i in range(n)], [0 for i in range(n)]
        b[0], c[0] = 1, 1
        l, m, i = 0, -1, 0
        while i < n:
            v = bs[(i - l):i]
            v = v[::-1]
            cc = c[1:l + 1]
            d = (bs[i] + sum(map(operator.mul, v, cc))) % 2
            if d == 1:
                tmp = [x for x in c]
                p = [0 for k in range(n)]
                for j in range(0, l):
                    if b[j] == 1:
                        p[j + i - m] = 1
                for k in range(len(c)):
                    c[k] = (c[k] + p[k]) % 2
                if l <= .5 * i:
                    l, m, b = i + 1 - l, i, tmp
            i += 1
        c = [i for i, x in enumerate(c) if x == 1]
        c.remove(0)
        self.seed, self.taps, self.n = LFSR._format_param(self.target[:l], c, l)
    
    def get(self, length=None, output_format="str", update=False):
        """
        Get a given number of bits from the LFSR with the given output format.
        
        :param length:        number of bits to be generated
        :param output_format: output format (string, hexadecimal or binary)
        :param update:        update the state of the generator
        """
        if output_format not in ["str", "hex", "bin"]:
            raise ValueError("Bad output format")
        length = length or (1 if self.target is None else len(self.target))
        stream = [b for b in self.seed]
        # generate length - nbits (from the starting stream) bits
        for _ in range(length - self.n):
            bit = 0
            for tap in self.taps:
                bit ^= stream[-tap]
            stream.append(bit)
        # update state if required
        if update:
            self.seed = stream[-self.n:]
            self.target = [b for b in self.seed]
        # now format the output according to output_format
        if output_format == "str":
            return bin2str(stream)
        elif output_format == "hex":
            return bin2hex(stream)
        return stream
    
    def test(self, target=None):
        """ Test the current target result against an input target. """
        target = LFSR._format_target(target or self.target)
        if target != self.get(len(target), "bin", False):
            raise ValueError("Target and generated bits do not match")
    
    @staticmethod
    def _format_param(seed, taps, nbits):
        """ Ensure that:
            - seed is formatted as a list of bits
            - taps is a list of integers <= nbits
            - nbits is an integer """
        if not is_pos_int(nbits, False):
            raise ValueError("Bad number of bits ({}) ; should be a non-null positive integer".format(nbits))
        if not is_list(taps) or any(not is_pos_int(i, False) or i > nbits for i in taps):
            raise ValueError("Bad taps ({}) ; should be a list of integers belonging to [1,nbits]".format(taps))
        if isinstance(seed, int):
            seed = int2bin(seed, nbits_in=nbits, nbits_out=nbits)
        elif is_str(seed) and not is_bin(seed):
            if is_hex(seed):
                seed = hex2str(seed)
            seed = str2bin(seed).zfill(nbits)
        if not is_bin(seed):
            raise ValueError("Bad seed ({}) ; should be a list of bits".format(seed))
        elif bin2int(seed) == 0:
            raise ValueError("Bad seed ; should not be null")
        elif len(seed) != nbits:
            raise ValueError("Bad seed ({} bits) ; should be {} bits".format(len(seed), nbits))
        return [int(b) for b in seed], taps, nbits
    
    @staticmethod
    def _format_target(target):
        """ Ensure that target is formatted as a list of bits """
        if is_str(target) and not is_bin(target):
            if is_hex(target):
                target = hex2str(target)
            target = [int(b) for b in str2bin(target)]
        if not is_bin(target):
            raise ValueError("Bad target ({}) ; should be a list of bits".format(target))
        return target


random.Geffe = Geffe
random.LFSR  = LFSR




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Module for defining benchmark mode logic.

"""
import signal
import time
from errno import ETIME
from os import strerror

from .helpers.timeout import TimeoutError


__all__ = ["set_time_items"]

TO_MSG = strerror(ETIME)


def set_time_items(glob):
    """
    This function prepares the benchmark items for inclusion in main script's
     global scope.
    
    :param glob: main script's global scope dictionary reference
    """
    a = glob['args']
    l = glob['logger']
    # Time manager, for keeping track of collected times
    class __TimeManager(object):
        """ Simple time manager, using time module. """
        def __init__(self):
            c = a._collisions
            self._stats = getattr(a, c.get("stats") or "stats", False)
            self._timings = getattr(a, c.get("timings") or "timings", False)
            self.enabled = self._stats or self._timings
            self.last = self.start = time.time()
            self.times = []
        
        def stats(self):
            end = time.time()
            b = ""
            for d, s, e in self.times:
                b += "\n{}\n> {} seconds".format(d, e - s)
            l.time("Total time: {} seconds{}".format(end - self.start, b))
    glob['time_manager'] = manager = __TimeManager()
    # private function to keep time measure in the time manager
    def _take_time(start=None, descr=None):
        t = manager.last = time.time()
        if start is not None and descr is not None:
            manager.times.append((descr, float(start), float(t)))
        return t - (start or 0)
    # Time context manager, for easilly benchmarking a block of code
    class Timer(object):
        def __init__(self, description=None, message=TO_MSG, timeout=None,
                     fail_on_timeout=False):
            self.fail = fail_on_timeout
            self.id = len(manager.times)
            self.descr = "#" + str(self.id) + \
                         (": " + (description or "")).rstrip(": ")
            self.message = message
            self.start = _take_time()
            self.timeout = timeout

        def __enter__(self):
            if manager.enabled:
                if self.timeout is not None:
                    signal.signal(signal.SIGALRM, self._handler)
                    signal.alarm(self.timeout)
                if manager._timings and self.descr:
                    l.time(self.descr)
                return self
        
        def __exit__(self, exc_type, exc_value, exc_traceback):
            if manager.enabled:
                d = _take_time(self.start, self.descr)
                if manager._timings:
                    l.time("> Time elapsed: {} seconds".format(d))
                if self.timeout is not None:
                    if not self.fail and exc_type is TimeoutError:
                        return True  # this allows to let the execution continue
            # implicitely returns None ; this lets the exception be raised
        
        def _handler(self, signum, frame):
            raise TimeoutError(self.message)
    
    glob['Timer'] = Timer
    # timing function for getting a measure from the start of the execution
    def get_time(message=None, start=manager.start):
        if manager._timings:
            l.time("> {}: {} seconds".format(message or "Time elapsed since "
                                          "execution start", _take_time(start)))
    glob['get_time'] = get_time
    # timing function for getting a measure since the last one
    def get_time_since_last(message=None):
        get_time(message or "Time elapsed since last measure", manager.last)
    glob['get_time_since_last'] = get_time_since_last



# -*- coding: UTF-8 -*-
"""Decorators for classes and methods.

"""
import logging
from functools import wraps
from sys import exc_info
try:  # PYTHON3
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec


__all__ = __features__ = ["applicable_to", "try_or_die", "try_and_pass", "try_and_warn"]


class IncompatibleClassError(Exception):
    pass


def __is_method(f):
    spec = getfullargspec(f)
    return len(spec.args) > 0 and spec.args[0] == "self"



def applicable_to(*classes):
    """
    Class decorator for checking that a class is well inherited from any given parent class. Used in checking mixins'
     compatibility.

    :param classes: list of compatible parent bot classes
    """
    def _wrapper(cls):
        class NewClass(cls):
            def __init__(self, *args, **kwargs):
                self.__doc__ = cls.__doc__
                self.__name__ = cls.__name__
                valid = [c.__name__ for c in classes]
                if all(c not in self.__class__.__mro__ for c in classes):
                    msg = "This class is not compatible with the given parent classes ({})".format(", ".join(valid))
                    raise IncompatibleClassError(msg)
                super(NewClass, self).__init__(*args, **kwargs)
        return NewClass
    return _wrapper


def try_or_die(message, exc=Exception, trace=True, extra_info=""):
    """
    Decorator handling a try-except block to log an error.

    :param message:    message to be displayed when crashing
    :param exc:        exception class on which the error is thrown
    :param trace:      display exception traceback
    :param extra_info: class attribute name whose value is to be displayed as additional information (only applies when
                        used with a method)
    """
    def _try_or_die(f):
        @wraps(f)
        @logging.bindLogger
        def wrapper(*args, **kwargs):
            self = args[0] if len(args) > 0 and __is_method(f) else None
            try:
                return f(*args, **kwargs)
            except exc as e:
                l = logger if self is None else self.logger
                l.critical(message)
                if trace:
                    l.exception(e)
                if extra_info != "" and hasattr(self, extra_info):
                    l.info(getattr(self, extra_info))
                # if the decorated method is part of a context manager, close it with its __exit__ method and continue
                if hasattr(self, "__exit__"):
                    self.__exit__(*exc_info())
                # re-raise the exception
                raise e
        return wrapper
    return _try_or_die


def try_and_pass(exc=Exception):
    """
    Decorator handling a try-except block that simply continue the execution with no message in case of failure.

    :param exc: exception class on which the error is thrown
    """
    def _try_and_pass(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except exc:
                pass
        return wrapper
    return _try_and_pass


def try_and_warn(message, exc=Exception, trace=False, extra_info=""):
    """
    Decorator handling a try-except block to log a warning and continue the execution.

    :param message:    message to be displayed when crashing
    :param exc:        exception class on which the error is thrown
    :param trace:      display exception traceback
    :param extra_info: class attribute name whose value is to be displayed as additional information
    """
    def _try_and_warn(f):
        @wraps(f)
        @logging.bindLogger
        def wrapper(*args, **kwargs):
            self = args[0] if len(args) > 0 and __is_method(f) else None
            try:
                return f(*args, **kwargs)
            except exc as e:
                l = logger if self is None else self.logger
                l.warning(message)
                if trace:
                    l.exception(e)
                if extra_info != "" and hasattr(self, extra_info):
                    l.info(getattr(self, extra_info))
        return wrapper
    return _try_and_warn




# -*- coding: UTF-8 -*-
"""Subpackage containing data transformation functions.

"""
from .common import *
from .report import *

from .common import __features__ as _common
from .report import __features__ as _report


__all__ = __features__ = _common + _report



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Module for defining argument parser-related functions and objects.

"""
import argparse
import atexit
import os
import re
import sys
from asciistuff import AsciiFile, Banner
from inspect import currentframe, getmembers, isfunction, ismethod
from os.path import basename, splitext
from six import string_types

from .argreparse import *
from .timing import set_time_items
from .handlers import *
from .helpers.constants import LINUX, PYTHON3
from .helpers.data.types import ip_address, port_number
from .helpers.text import configure_docformat, gt
from .interact import set_interact_items
from .hotkeys import set_hotkeys
from .notify import set_notify_items
from .progress import set_progress_items
from .loglib import *
from .step import set_step_items


__all__ = __features__ = ["parser", "initialize", "validate"]

AT_EXIT_SET = False
BANNER_FONT = None
BANNER_STYLE = {}

parser_calls = []  # will be populated by calls to ProxyArgumentParser


def _save_config(glob):
    cf = glob['args'].write_config
    if cf:
        with open(cf, 'w') as f:
            glob['parser']._config.write(f)
        glob['logger'].debug(gt("Input arguments written to file '{}'").format(cf))


def initialize(add_banner=False,
               add_config=False,
               add_demo=False,
               add_interact=False,
               add_notify=False,
               add_progress=False,
               add_step=False,
               add_time=False,
               add_version=False,
               add_wizard=False,
               sudo=False,
               multi_level_debug=False,
               short_long_help=True,
               action_at_interrupt="exit",
               ext_logging=False,
               noargs_action=None,
               post_actions=True,
               report_func=None,
               **kwargs):
    """
    Initialization function ; sets up the arguments for the parser and creates a logger to be inserted in the input
     dictionary of global variables from the calling script.

    :param add_banner:          add an ASCII banner when starting the tool
    :param add_config:          add an option to input an INI configuration file
    :param add_demo:            add an option to re-run the process using a
                                 random entry from the __examples__ (only works
                                 if this variable is populated)
    :param add_interact:        add an interaction option
    :param add_notify:          add a notification option
    :param add_progress:        add a progress management option
    :param add_step:            add an execution stepping option
    :param add_time:            add an execution timing option
    :param add_version:         add a version option
    :param add_wizard:          add an option to run a wizard, asking for each
                                 input argument
    :param sudo:                if True, require sudo credentials and re-run
                                 script with sudo
    :param multi_level_debug:   allow to use -v, -vv, -vvv (adjust logging
                                 level) instead of just -v (only debug on/off)
    :param short_long_help:     enable/disable the separation of -h/--help
    :param action_at_interrupt: perform an action at interrupt
                                 (confirm|continue|exit)
    :param ext_logging:         extended logging options
    :param noargs_action:       action to be performed when no argument is input
    :param post_actions:        enable post-actions at interrupt
    :param report_func:         report generation function
    """
    global parser, parser_calls
    # handle backward-compatibility arguments
    exit_at_interrupt = kwargs.pop('exit_at_interrupt', None)
    if len(kwargs) > 0:
        raise TypeError("Unexpected keyword-argument{} ({})"
                        .format(["", "s"][len(kwargs) > 1], ", ".join(kwargs.keys())))
    # get caller's frame
    frame = currentframe().f_back
    # walk the stack until a frame containing a known object is found
    glob = {}
    while frame:
        if isinstance(frame.f_globals.get('parser'), ProxyArgumentParser):
            glob = frame.f_globals
            # search for dunders
            for d in DUNDERS:
                f = frame
                while f and (d not in f.f_globals.keys() or f.f_globals[d] is None):
                    f = f.f_back
                try:
                    glob[d] = f.f_globals[d]
                except (AttributeError, KeyError):
                    pass
            break
        frame = frame.f_back
    if any(glob.get(k) is not None for k in ["NOTIFICATION_ICONS_PATH", "NOTIFICATION_LEVEL", "NOTIFICATION_TIMEOUT"]):
        add_notify = True
    add = {'config': add_config, 'demo': add_demo, 'interact': add_interact, 'notify': add_notify,
           'progress': add_progress, 'step': add_step, 'time': add_time, 'version': add_version, 'wizard': add_wizard,
           'help': True, 'usage': True}
    p = ArgumentParser(glob)
    # 1) handle action when no input argument is given
    add['demo'] = add['demo'] and p.examples
    noarg = len(sys.argv) == 1
    if noarg and noargs_action:
        if noargs_action not in add.keys():
            raise ValueError(gt("Bad action when no args (should be one of: {})").format('|'.join(add.keys())))
        add[noargs_action] = True  # ensure this action is enabled, even if it is not given the passed arguments
    # 2) populate the real parser and add information arguments
    __parsers = {parser: p}
    #  proxy parser to real parser recursive conversion function
    def __proxy_to_real_parser(value):
        """
        This recursively converts ProxyArgumentParser instances to real parsers.

        Use case: defining subparsers with a parent
          >>> [...]
          >>> parser.add_argument(...)  # argument common to all subparsers
          >>> subparsers = parser.add_subparsers()
          >>> subparsers.add_parser(..., parents=[parent])
                                                    ^
                   this is an instance of ProxyArgumentParser and must be converted to an actual parser instance

        :param value: a value coming from args or kwargs aimed to a real parser
        """
        if isinstance(value, ProxyArgumentParser):
            return __parsers[value]
        elif isinstance(value, (list, tuple)):
            return [__proxy_to_real_parser(_) for _ in value]
        return value
    #  now iterate over the registered calls
    for proxy_parser, method, args, kwargs, proxy_subparser in parser_calls:
        real_parser = __parsers[proxy_parser]
        args = (__proxy_to_real_parser(v) for v in args)
        kwargs = {k: __proxy_to_real_parser(v) for k, v in kwargs.items()}
        real_subparser = getattr(real_parser, method)(*args, **kwargs)
        if real_subparser is not None:
            __parsers[proxy_subparser] = real_subparser
    # this allows to ensure that another call to initialize(...) will have a clean list of calls and an empty _config
    #  attribute
    parser_calls = []
    ArgumentParser.reset()
    # configure documentation formatting
    configure_docformat(glob)
    # config handling feature, for reading/writing an INI config file with the input arguments, e.g. for future reuse
    if add['config']:
        c = p.add_argument_group("config arguments")
        opt = c.add_argument("-r", "--read-config", action="config", help=gt("read args from a config file"),
                             note=gt("this overrides other arguments"))
        c.add_argument("-w", "--write-config", metavar="INI", help=gt("write args to a config file"))
        if noarg and noargs_action == "config":
            sys.argv[1:] = [opt, "config.ini"]
    i = p.add_argument_group("extra arguments")
    # demonstration feature, for executing an example amongst these defined in __examples__, useful for observing what
    #  the tool does
    if add['demo']:
        opt = i.add_argument("--demo", action="demo", prefix="play", help=gt("demonstrate a random example"))
        if noarg and noargs_action == "demo":
            sys.argv[1:] = [opt]
    # help feature, for displaying classical or extended help about the tool
    if add['help']:
        if glob.get('__details__'):
            opt = i.add_argument("-h", dest="help", default=0, help=gt("show extended help message and exit"),
                                 action="count", note=gt("-hhh is the highest help detail level"))
        elif short_long_help:
            opt = i.add_argument("-h", dest="usage", action="usage", help=gt("show usage message and exit"))
            opt = i.add_argument("--help", action="help", help=gt("show this help message and exit"))
        else:
            opt = i.add_argument("-h", "--help", action="help", help=gt("show this help message and exit"))
        if noarg and noargs_action == "help":
            sys.argv[1:] = ["--help"]
        elif noarg and noargs_action == "usage":
            sys.argv[1:] = ["DISPLAY_USAGE"]
    # interaction mode feature, for interacting with the tool during its execution, useful for debugging
    if add['interact']:
        j = p.add_argument_group("interaction arguments")
        opt = j.add_argument("--interact", action="store_true", suffix="mode", help=gt("interaction mode"))
        if opt:
            j.add_argument("--host", default="127.0.0.1", type=ip_address, prefix="remote",
                           help=gt("remote interacting host"))
            j.add_argument("--port", default=12345, type=port_number, prefix="remote",
                           help=gt("remote interacting port"))
        if noarg and noargs_action == "interact":
            sys.argv[1:] = [opt]
    # notification feature, for displaying notifications during the execution
    if add['notify']:
        opt = i.add_argument("-n", "--notify", action="store_true", suffix="mode", help=gt("notify mode"))
        if noarg and noargs_action == "notify":
            sys.argv[1:] = [opt]
    # progress mode feature, for displaying a progress bar during the execution
    if add['progress']:
        opt = i.add_argument("-p", "--progress", action="store_true", suffix="mode", help=gt("progress mode"))
        if noarg and noargs_action == "progress":
            sys.argv[1:] = [opt]
    # stepping mode feature, for stepping within the tool during its execution, especially useful for debugging
    if add['step']:
        opt = i.add_argument("--step", action="store_true", last=True, suffix="mode", help=gt("stepping mode"))
        if noarg and noargs_action == "step":
            sys.argv[1:] = [opt]
    # timing mode feature, for measuring time along the execution of the tool
    if add['time']:
        b = p.add_argument_group("timing arguments")
        opt = b.add_argument("--stats", action='store_true', last=True, prefix="time",
                             help=gt("display execution time stats at exit"))
        b.add_argument("--timings", action='store_true', last=True, suffix="mode",
                       help=gt("display time stats during execution"))
        if noarg and noargs_action == "time":
            sys.argv[1:] = [opt]
    # version feature, for displaying the version from __version__
    if add['version']:
        version = glob['__version__'] if '__version__' in glob else None
        if version is not None:
            opt = i.add_argument("--version", action='version', prefix="show", version=version,
                                 help=gt("show program's version number and exit"))
            if noarg and noargs_action == "version":
                sys.argv[1:] = [opt]
    # verbosity feature, for displaying debugging messages, with the possibility to handle multi-level verbosity
    if multi_level_debug:
        i.add_argument("-v", dest="verbose", default=0, action="count", suffix="mode", cancel=True, last=True,
                       help=gt("verbose level"), note=gt("-vvv is the highest verbosity level"))
    else:
        i.add_argument("-v", "--verbose", action="store_true", last=True, suffix="mode", help=gt("verbose mode"))
    # wizard feature, for asking argument values to the user
    if add['wizard']:
        opt = i.add_argument("-w", "--wizard", action="wizard", prefix="start", help=gt("start a wizard"))
        if noarg and noargs_action == "wizard":
            sys.argv[1:] = [opt]
    # reporting feature, for making a reporting with the results of the tool at the end of its execution
    if report_func is not None and PYTHON3:
        if not isfunction(report_func):
            report_func = None
            glob['logger'].error(gt("Bad report generation function"))
        else:
            # lazily import report features
            #  -> reason: they rely on weasyprint, which take time to be imported ; so, when report features are not
            #              used in a script, report classes won't be loaded
            all_list = __import__("tinyscript.report", fromlist=['__all__']).__all__
            report = __import__("tinyscript.report", fromlist=all_list)
            for f in all_list:
                glob[f] = globals()[f] = getattr(report, f)
            # now populate the parser with report-related arguments
            r = p.add_argument_group("report arguments")
            output_func = list(filter(lambda x: getattr(x[1], '_output', False), getmembers(Report)))
            choices = list(map(lambda x: x[0], output_func))
            if r.add_argument("--output", choices=choices, default="pdf", last=True, prefix="report",
                              help=gt("report output format")):
                r.add_argument("--title", last=True, prefix="report", help=gt("report title"))
                r.add_argument("--css", last=True, prefix="report", help=gt("report stylesheet file"))
                r.add_argument("--theme", default="default", last=True, prefix="report",
                               help=gt("report stylesheet theme"), note=gt("--css overrides this setting"))
                r.add_argument("--filename", last=True, prefix="report", help=gt("report filename"))
    elif report_func is not None and not PYTHON3:
        report_func = None  # disable reporting in the at_exit handler
        glob['logger'].warning(gt("Report generation is only available with Python 3"))
    # extended logging features
    if ext_logging:
        i.add_argument("-f", "--logfile", last=True, help=gt("destination log file"))
        i.add_argument("-r", "--relative", action="store_true", last=True, suffix="time",
                       help=gt("display relative time"))
        if LINUX:
            i.add_argument("-s", "--syslog", action="store_true", last=True, suffix="mode",
                           help=gt("log to /var/log/syslog"))
    # now parse inputs
    glob['args'], glob['parser'] = p.parse_args(), p
    # 3) if sudo required, restart the script
    if sudo:
        # if not root, restart the script in another process and jump to this
        if os.geteuid() != 0:
            os.execvp("sudo", ["sudo", sys.executable] + sys.argv)
    # 4) configure logging and get the main logger
    configure_logger(glob, multi_level_debug, glob['args']._collisions.get("relative"),
                     glob['args']._collisions.get("logfile"), glob['args']._collisions.get("syslog"))
    # 5) append modes items
    set_interact_items(glob)
    set_hotkeys(glob)
    set_notify_items(glob)
    set_progress_items(glob)
    set_step_items(glob)
    set_time_items(glob)
    # 6) display a banner if relevant
    bf = glob.get('BANNER_FONT', BANNER_FONT)
    if add_banner or isinstance(bf, string_types):
        f = AsciiFile()
        bs = BANNER_STYLE
        bs = bs if isinstance(bs, dict) else {}
        f['title', bs] = Banner(p.scriptname, font=bf)
        print(f)
    # 7) finally, bind the global exit handler
    _hooks.sigint_action = action_at_interrupt if exit_at_interrupt is None else ["continue", "exit"][exit_at_interrupt]
    def __at_exit():
        # first, dump the config if required
        if add['config']:
            _save_config(glob)
        # now, close the current progress bar (if any)
        if add['progress']:
            glob['progress_manager'].stop()
        # then handle the state
        do_post_actions = True
        if _hooks.state == "INTERRUPTED":
            glob.get('at_interrupt', lambda: None)()
            do_post_actions = post_actions
        elif _hooks.state == "TERMINATED":
            glob.get('at_terminate', lambda: None)()
            do_post_actions = False
        # finally handle post-actions
        if do_post_actions:
            if report_func is not None:
                # generate the report only when exiting gracefully, just before at_graceful_exit
                a = glob['args']
                try:
                    r = Report(*report_func(), title=a.title, filename=a.filename, css=a.css, logger=glob['logger'])
                    getattr(r, a.output)(False)
                except AttributeError:
                    pass
            t = glob['time_manager']
            if add['time'] and t._stats:
                t.stats()
            glob.get('at_graceful_exit', lambda: None)()
        glob.get('at_exit', lambda: None)()
        from logging import shutdown
        shutdown()
    if not AT_EXIT_SET:
        atexit.register(__at_exit)
        globals()['AT_EXIT_SET'] = False


def validate(*arg_checks):
    """
    Function for validating group of arguments ; each argument is represented as a 4-tuple like follows:

        (argument_name, fail_condition, error_message, default)

        - argument_name: the name of the argument like entered in add_argument()
        - fail_condition: condition in Python code with the argument name replaced by ' ? ' (e.g. " ? > 0")
        - error_message: message describing what failed with the validation ofs the current argument
        - default [optional]: value to be given if the validation fails ; this implies that the script will not exit
                              after the validation (if no other blocking argument present)

    :param arg_checks: list of 3/4-tuples
    """
    glob = currentframe().f_back.f_globals
    if glob.get('args') is None:
        return
    locals().update(glob)
    exit_app = False
    for check in arg_checks:
        check = check + (None, ) * (4 - len(check))
        param, condition, message, default = check
        if re.match(r'^_?[a-zA-Z][a-zA-Z0-9_]*$', param) is None:
            raise ValueError(gt("Illegal argument name"))
        if not hasattr(glob['args'], param):
            raise AttributeError("'{}' argument does not exist".format(param))
        try:
            result = eval(condition.replace(" ? ", " glob['args'].{} ".format(param)))
        except Exception as e:
            result = True
            message = str(e)
        if result:
            if default is None:
                logger.error(gt(message or "Validation failed"))
                exit_app = True
            else:
                logger.warning(gt(message or "Validation failed"))
                setattr(glob['args'], param, default)
    if exit_app:
        sys.exit(2)


class ProxyArgumentParser(object):
    """
    Proxy class for collecting added arguments before initialization.
    """
    def __getattr__(self, name):
        """ Each time a method is called, return __collect to make it capture the input arguments and keyword-arguments
             if it exists in the original parser class. """
        self.__call = name
        return self.__collect

    def __collect(self, *args, **kwargs):
        """ Capture the input arguments and keyword-arguments of the currently called method, appending a proxy
             subparser in case it should be used for mutually exclusive groups or subparsers. """
        subparser = ProxyArgumentParser()
        parser_calls.append((self, self.__call, args, kwargs, subparser))
        del self.__call
        return subparser


parser = ProxyArgumentParser()



# -*- coding: UTF-8 -*-
"""Common execution functions and decorators.

"""
from functools import wraps
from multiprocessing import Process
from six import string_types
from subprocess import Popen, PIPE
from threading import Thread


__all__ = __features__ = ["execute", "process", "processes_clean", "thread", "threads_clean"]


PROCESSES = []
THREADS   = []


def execute(cmd, **kwargs):
    """
    Dummy wrapper for subprocess.Popen.

    :param cmd: command string
    """
    if isinstance(cmd, string_types):
        cmd = cmd.split()
    return Popen(cmd, stdout=PIPE, stderr=PIPE, **kwargs).communicate()


def process(f):
    """
    Decorator function for simply running the given function as a separate process.

    :param f: function to be run as a separate process
    """
    @wraps(f)
    def _wrapper(*args, **kwargs):
        p = Process(target=f, args=args, kwargs=kwargs)
        p.start()
        PROCESSES.append(p)
        return p
    return _wrapper


def processes_clean(terminate=False, timeout=None):
    """
    Utility function to clean up the list of processes.

    :return:
    """
    global PROCESSES
    for p in PROCESSES:
        p.terminate() if terminate else p.join(timeout)
    PROCESSES = []


def thread(f):
    """
    Decorator function for simply running the given function as a separate thread.

    :param f: function to be run as a separate thread
    """
    @wraps(f)
    def _wrapper(*args, **kwargs):
        t = Thread(target=f, args=args, kwargs=kwargs)
        t.start()
        THREADS.append(t)
        return t
    return _wrapper


def threads_clean(timeout=None):
    """
    Utility function to clean up the list of threads.

    :return:
    """
    global THREADS
    for t in THREADS:
        t.join(timeout)
    THREADS = []




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Module for defining progress mode logic.

"""
from tqdm import tqdm, trange


__all__ = ["set_progress_items"]


def set_progress_items(glob):
    """
    This function prepares the progress items for inclusion in main script's global scope.
    
    :param glob: main script's global scope dictionary reference
    """
    a = glob['args']
    enabled = getattr(a, a._collisions.get("progress") or "progress", False)
    # Progress manager, for providing an interface to tqdm progress bar class
    class __ProgressManager(object):
        """ Simple progress bar manager, relying on tqdm module. """
        def __init__(self):
            c = a._collisions
            self._tqdm = None
        
        def __getattr__(self, name):
            if enabled:
                try:
                    return self.__getattribute__(name)
                except AttributeError:
                    pass
                if hasattr(tqdm, name) and self._tqdm is not None:
                    return getattr(self._tqdm, name)
                raise AttributeError("ProgressManager instance has no attribute '{}'".format(name))
        
        def range(self, *args, **kwargs):
            """ Dummy alias to trange. """
            if enabled:
                self._tqdm = trange(*args, **kwargs)
                return self._tqdm
        
        def start(self, *args, **kwargs):
            if enabled:
                self.stop()
                self._tqdm = tqdm(*args, **kwargs)
                return self._tqdm
        
        def stop(self):
            """ Closing method. """
            if enabled:
                if self._tqdm is not None:
                    self._tqdm.close()
                self._tqdm = None
    glob['progress_manager'] = manager = __ProgressManager()
    # shortcut function to range-based progress bar
    def progressbar(*args, **kwargs):
        """ Range-based progress bar relying on tqdm. """
        try:
            iter(args[0])
            return manager.start(*args, **kwargs)
        except TypeError as te:
            return manager.range(*args, **kwargs)
    glob['progressbar'] = progressbar
